from __future__ import annotations

import copy
import inspect
import pydoc
import re
import types
import typing as t
import webbrowser
from pathlib import Path

from pyqtgraph.Qt import QtCore

from s3a import PRJ_ENUMS
from s3a.constants import MENU_OPTS_DIR
from s3a.shims import typing_extensions as t_e
from utilitys import NestedProcess, RunOpts, ProcessIO
from utilitys import ParamEditor, NestedProcWrapper, fns, ProcessStage, AtomicProcess
from utilitys.typeoverloads import FilePath
from utilitys.widgets import makeDummySignal

Signal = QtCore.Signal
_procDict = t.Dict[str, t.List[t.Union[str, dict]]]

class _AlgClctnDict(t_e.TypedDict):
  top: _procDict
  primitive: _procDict
  modules: t.List[str]

_underscoreMatcher = re.compile(r'([A-Za-z])_+([A-Za-z])')
def _fmt(name: str):
  # Don't replace first/last underscore to avoid publicizing private / utility functions
  name = name.lower().replace(' ', '')
  return re.sub(_underscoreMatcher, r'\1\2', name)

algoNameFormatter = fns.NameFormatter(_fmt)
"""Strips spaces and underscores from the provided name, and turns everything to 
lowercase"""

class AlgCollection(ParamEditor):
  # sigProcessorCreated = Signal(object) # Signal(AlgCollectionEditor)
  def __init__(self, procWrapType=NestedProcWrapper, procType=NestedProcess, parent=None, fileType='alg',
               template: FilePath=None, **kwargs):
    super().__init__(parent, fileType=fileType, **kwargs)
    self.procWrapType = procWrapType
    self.procType = procType
    self.primitiveProcs: t.Dict[str, t.Union[ProcessStage, t.List[str]]] = {}
    self.topProcs: _procDict = {}
    self.includeModules: t.List[str] = []

    if template is not None:
      self.loadParamValues(template)

    if not self.topProcs:
      # Ensure at least one top-level processor exists
      self.addProcess(self.procType('None'), top=True)

    # Make sure fallthroughs are possible, i.e. for really simple image processes
    self.addProcess(AtomicProcess(lambda **_kwargs: ProcessIO(**_kwargs), name='Fallthrough'))

  def createProcessorEditor(self, saveDir: t.Union[str, t.Type], editorName='Processor') -> AlgParamEditor:
    """
    Creates a processor editor capable of dynamically loading, saving, and editing collection processes

    :param saveDir: Directory for saved states. If a class type is provided, the __name__ of this class is used.
      Note: Either way, the resulting name is lowercased before being applied.
    :param editorName: The name of the spawned editor
    """
    if not isinstance(saveDir, str):
      saveDir = saveDir.__name__
    formattedClsName = fns.pascalCaseToTitle(saveDir)
    editorDir = MENU_OPTS_DIR/formattedClsName.lower()
    return AlgParamEditor(self, saveDir=editorDir, fileType=self.fileType, name=editorName)

  def addProcess(self, proc: ProcessStage, top=False, force=False):

    addDict = self.topProcs if top else self.primitiveProcs
    saveObj = {proc.name: proc} if isinstance(proc, AtomicProcess) else proc.saveState_flattened()
    if force or proc.name not in addDict:
      addDict.update(saveObj)
    for stage in proc:
      # Don't recurse 'top', since it should only hold the directly passed process
      self.addProcess(stage, False, force)
    return proc.name

  def addFunction(self, func: t.Callable, top=False, **kwargs):
    """Helper function to wrap a function in an atomic process and add it as a stage"""
    return self.addProcess(AtomicProcess(func, **kwargs), top)

  def parseProcStages(self, stages: t.Sequence[t.Union[dict, str]], name:str=None, add=PRJ_ENUMS.PROC_NO_ADD,
                      allowOverwrite=False):
    """
    Creates a nested process from a sequence of process stages and optional name
    :param stages: Stages to parse
    :param name: Name of the nested process
    :param add: Whether to add this new process to the current collection's top or primitive process blocks, or
      to not add at all (if NO_ADD)
    :param allowOverwrite: If `add` is *True*, this determines whether the new process can overwite an already existing
      proess. If `add` is *False*, this value is ignored.
    """
    out = self.procType(name)
    for stageName in stages:
      if isinstance(stageName, dict):
        stage = self.parseProcDict(stageName)
      else:
        stage = self.parseProcName(stageName, topFirst=False)
      out.addProcess(stage)
    exists = out.name in self.topProcs
    if add is not PRJ_ENUMS.PROC_NO_ADD and (not exists or allowOverwrite):
      self.addProcess(out, top=add is PRJ_ENUMS.PROC_ADD_TOP, force=allowOverwrite)
    return out

  def parseProcName(self, procName: str, topFirst=True, searchDicts: t.Sequence[dict]=None, **kwargs):
    """
    From a list of search locations (ranging from most primitive to topmost), find the first processor matching the
    specified name. If 'topFirst' is chosen, the search locations are parsed in reverse order.
    """
    if searchDicts is None:
      searchDicts = [self.primitiveProcs, self.topProcs]
    if topFirst:
      searchDicts = searchDicts[::-1]
    proc = searchDicts[0].get(procName, searchDicts[1].get(procName))
    # It could still be in an include module, cache if found
    if not proc:
      for module in self.includeModules:
        proc = self.parseProcModule(module, procName, **kwargs)
        if proc:
          # Success, make sure to cache this in processes
          # Top processes must be nested
          self.addProcess(proc, topFirst and isinstance(proc, NestedProcess))
          break
    if proc is None:
      raise ValueError(f'Process "{procName}" not recognized')
    if not isinstance(proc, ProcessStage):
      proc = self.parseProcStages(proc, procName)
    else:
      proc = copy.deepcopy(proc)
      # Default to disableale stages. For non-disablable, use parseDict
      proc.allowDisable = True
    return proc

  def parseProcDict(self, procDict: dict, topFirst=False):
    # 1. First key is always the process name, values are new inputs for any matching process
    # 2. Second key is whether the process is disabled
    procName, updateArgs = next(iter(procDict.items()))
    procDict.pop(procName)
    proc = self.parseProcName(procName, topFirst=topFirst)
    if isinstance(updateArgs, list):
      # TODO: Determine policy for loading nested procs, outer should already know about inner so it shouldn't
      #   _need_ to occur, given outer would've been saved previously
      raise ValueError('Parsing deep nested processes is currently undefined')
    elif updateArgs:
      proc.updateInput(**updateArgs, allowExtra=True, graceful=True)
      # Set the defaults, too
      # if isinstance(proc, AtomicProcess):
      #   proc.defaultInput.update(**updateArgs)
    # Check for process-level traits
    for kk, vv in procDict.items():
      if hasattr(proc, kk):
        setattr(proc, kk, vv)
    return proc
    # TODO: Add recursion, if it's something that will be done. For now, assume only 1-depth nesting. Otherwise,
    #   it's hard to distinguish between actual input optiosn and a nested process

  def _addFromModuleName(self, fullModuleName: str, primitive=True):
    """
    Adds all processes defined in a module to this collection. From a full module name (import.path.module). Rules:
      - All functions defined in that file *not* beginning with an underscore (_) will be added, except for
        the rule(s) below
      - All functions ending with 'factory' will be assumed ProcessStage factories, where their return value is exactly
        one ProcessStage. These are expected to take no arguments. Note that if `primitive` is *False* and an
        AtomicProcess is returned, errors will occur. So, it is implicitly forced to be a primitive process if this
        occurs

    :param fullModuleName: Module name to parse. Should be in a format expected by pydoc
    :param primitive: Whether the returned values should be considered top or primitive processes
    """
    if fullModuleName in self.includeModules:
      return
    module = pydoc.locate(fullModuleName)
    if not module:
      raise ValueError(f'Module "{fullModuleName}" not recognized')
    for name, func in inspect.getmembers(module,
                                         lambda el: inspect.isfunction(el)
                                                    and el.__module__ == module.__name__
                                                    and not el.__name__.startswith('_')):
      if name.lower().endswith('factory'):
        obj = func()
        self.addProcess(obj, top=not primitive and not isinstance(obj, AtomicProcess))
      else:
        self.addFunction(func)

  @classmethod
  def parseProcModule(cls, moduleName: t.Union[str, types.ModuleType], procName: str,
                                       formatter=algoNameFormatter, **factoryArgs):
    if isinstance(moduleName, types.ModuleType):
      module = moduleName
    else:
      module = pydoc.locate(moduleName)
    if not module:
      raise ValueError(f'Module "{moduleName}" not recognized')
    # TODO: Depending on search time, maybe quickly search without formatting?
    procName = formatter(procName)
    # It is possible after name formatting for multiple matches to exist for procName.
    # The first match that is a function or process stage will be retained.
    for name, attr in vars(module).items():
      name = formatter(name.split('.')[-1])
      # Evaluate factories first
      if name.lower().endswith('factory'):
        try:
          attr: ProcessStage = attr(**factoryArgs)
          if not isinstance(attr, ProcessStage):
            continue
          name = formatter(attr.name)
        except:
          continue
      if name == procName:
        if isinstance(attr, ProcessStage):
          return attr
        if callable(attr):
          return AtomicProcess(attr, **factoryArgs)
    return None

  def saveParamValues(self, saveName: str=None, paramState: dict=None, **kwargs):
    def procFilter(procDict):
      return {k: v for k, v in procDict.items() if not isinstance(v, ProcessStage)}
    if paramState is None:
      paramState = {'top': procFilter(self.topProcs), 'primitive': procFilter(self.primitiveProcs),
                    'modules': self.includeModules}
    return super().saveParamValues(saveName, paramState, **kwargs)

  def loadParamValues(self, stateName: t.Union[str, Path],
                      stateDict: _AlgClctnDict=None,
                      **kwargs):
    stateDict = self._parseStateDict(stateName, stateDict)
    top, primitive = stateDict.get('top', {}), stateDict.get('primitive', {})
    modules = stateDict.get('modules', [])
    self.includeModules = modules
    self.topProcs.update(top)
    self.primitiveProcs.update(primitive)

    return super().loadParamValues(stateName, stateDict, candidateParams=[])

class AlgParamEditor(ParamEditor):
  sigProcessorChanged = QtCore.Signal(str)
  """Name of newly selected process"""

  def __init__(self, clctn: AlgCollection=None, **kwargs):
    super().__init__(**kwargs)
    if clctn is None:
      clctn = AlgCollection()
    self.clctn = clctn
    self.treeBtnsWidget.hide()

    procName = next(iter(self.clctn.topProcs))
    proc = clctn.parseProcName(procName)
    # Set to None first to force switch, init states
    self.curProcessor = self.clctn.procWrapType(proc, self.params)
    _, self.changeProcParam = self.registerFunc(self.changeActiveProcessor, runOpts=RunOpts.ON_CHANGED, returnParam=True,
                                                overrideBasePath=(), parentParam=self._metaParamGrp,
                                                proc=procName)
    fns.setParamsExpanded(self._metaTree)
    procSelector = self.changeProcParam.child('proc')
    self.clctn.sigChangesApplied.connect(lambda: procSelector.setLimits(list(self.clctn.topProcs)))
    self.clctn.sigChangesApplied.emit({})
    def onChange(name):
      with makeDummySignal(procSelector, 'sigValueChanged'):
        procSelector.setValue(name)
        # Manually set item labels since valueChange was forcefully disconnected
        for item in procSelector.items:
          item.valueChanged(procSelector, name)
    self.sigProcessorChanged.connect(onChange)
    self.changeActiveProcessor(procName)

  def _unnestedProcState(self, proc: NestedProcess, _state=None, **kwargs) -> _AlgClctnDict:
    """
    Updates processes without hierarchy so separate stages are unnested. The outermost process is considered a
    'top' process, while all subprocesses are considered 'primitive'.

    :param proc: Process to record the unnested state
    :param _state: Internally used, do not provide in the function call. It will be returned at the end with 'top' and
      'primitive' keys
    :return: Mock Collection state from just the provided nested process. The passed process will be the only 'top'
      value, while all substages are entries (and unnested substages, etc.) are entries in the 'primitive' key
    """
    kwargs.update(includeMeta=True, disabled=False, allowDisable=True)
    first = _state is None
    if first:
      _state = {'modules': self.clctn.includeModules, 'primitive': {}, 'top': {}}
    stageVals = []
    for stage in proc:
      if isinstance(stage, NestedProcess):
        self._unnestedProcState(stage, _state, **kwargs)
        stageVals.append(stage.addMetaProps(stage.name, **kwargs))
      else:
        stageVals.append(stage.saveState(**kwargs))
    entryPt = 'top' if first else 'primitive'
    _state[entryPt][proc.name] = stageVals
    return _state

  def saveParamValues(self, saveName: str=None, paramState: dict=None, *, includeDefaults=False, **kwargs):
    """
    The algorithm editor also needs to store information about the selected algorithm, so lump
    this in with the other parameter information before calling default save.
    """
    proc = self.curProcessor.processor
    # Make sure any newly added stages are accounted for
    if paramState is None:
      # Since inner nested processes are already recorded, flatten here to just save updated parameter values for the
      # outermost stage
      paramState = self._unnestedProcState(proc, includeMeta=True)
      paramState['modules'] = self.clctn.includeModules
    self.clctn.loadParamValues(self.clctn.stateName, paramState)
    clctnState = self.clctn.saveParamValues(saveName, blockWrite=True)
    paramState = {'Selected Algorithm': self.curProcessor.algName, 'Parameters': clctnState}
    return super().saveParamValues(saveName, paramState, includeDefaults=includeDefaults, **kwargs)

  def loadParamValues(self, stateName: t.Union[str, Path],
                      stateDict: dict=None, **kwargs):
    stateDict = self._parseStateDict(stateName, stateDict)
    procName = stateDict.get('Selected Algorithm')
    if not procName:
      procName = next(iter(self.clctn.topProcs))
    clctnState = stateDict['Parameters']

    self.clctn.loadParamValues(stateName, clctnState, **kwargs)
    self.changeActiveProcessor(procName, saveBeforeChange=False)
    return super().loadParamValues(stateName, stateDict, candidateParams=[], **kwargs)

  def changeActiveProcessor(self, proc: t.Union[str, NestedProcess], saveBeforeChange=True):
    """
    Changes which processor is active.

    :param proc:
      helpText: Processor to load
      pType: popuplineeditor
      limits: []
      title: Algorithm
    :param saveBeforeChange: Whether to propagate current algorithm settings to the processor collection before changing
    """
    # TODO: Maybe there's a better way of doing this? Ensures proc label is updated for programmatic calls
    if saveBeforeChange:
      self.saveParamValues(self.stateName, self._unnestedProcState(self.curProcessor.processor, includeMeta=True),
                           blockWrite=True)
    if isinstance(proc, str):
      proc = self.clctn.parseProcName(proc)
    if proc == self.curProcessor.processor:
      return
    self.curProcessor.clear()
    self.params.clearChildren()
    self.curProcessor = self.clctn.procWrapType(proc, self.params)
    fns.setParamsExpanded(self.tree)
    self.sigProcessorChanged.emit(proc.name)

  def editParamValues_gui(self):
    webbrowser.open(self.formatFileName(self.stateName))


import inspect
from ast import literal_eval
import typing as t

import numpy as np
import pandas as pd
from s3a.structures import AnnParseError

from ..structures import PrjParamGroup, ComplexXYVertices, XYVertices
from utilitys import PrjParam

_serFunc = t.Callable
_deserFunc = t.Callable
_valsType: t.Sequence[t.Any]

def _newHandlerTemplate():
  return {
    'serialize': None,
    'deserialize': None,
    'takesParam': False
  }

_serdesHandlers = pd.DataFrame(columns=list(_newHandlerTemplate()))

def registerIoHandler(pType: str, force=False, **kwargs):
  global _serdesHandlers
  if pType not in _serdesHandlers.index:
    _serdesHandlers = _serdesHandlers.append(pd.Series(_newHandlerTemplate(), name=pType))
  for which in {'serialize', 'deserialize'} & set(kwargs):
    if _serdesHandlers.loc[pType, which] is not None and not force:
      raise KeyError(f'Already have {which} handler for "{pType}"')
  _serdesHandlers.loc[pType, list(kwargs)] = list(kwargs.values())

def _runFunc(param: PrjParam, values, which: str, default: t.Callable, returnErrs=True):
  out = {}
  errs = {}
  parseType = param.opts.get('parser', param.pType)
  if parseType not in _serdesHandlers.index:
    handlerRow = _newHandlerTemplate()
  else:
    handlerRow = _serdesHandlers.loc[parseType]
  _handler = handlerRow[which] or default
  takesParam = handlerRow['takesParam']
  for ii, val in enumerate(values):
    try:
      out[ii] = _handler(param, val) if takesParam else _handler(val)
    except Exception as ex:
      errs[ii] = ex
  out = pd.Series(out, name=param, dtype=object)
  if returnErrs:
    return out, pd.Series(errs, name=param, dtype=object)
  return out

def serialize(param: PrjParam, values: t.Sequence[t.Any], returnErrs=True):
  default = lambda val: str(val)
  return _runFunc(param, values, 'serialize', default, returnErrs)

def deserialize(param: PrjParam, values: t.Sequence[str], returnErrs=True):
  paramType = type(param.value)
  default = lambda val: paramType(val)
  return _runFunc(param, values, 'deserialize', default, returnErrs)


from __future__ import annotations

from typing import Callable, Sequence

import cv2 as cv
import numpy as np
import pandas as pd
from pyqtgraph import console as pg_console
from pyqtgraph.Qt import QtCore, QtWidgets, QtGui

from s3a.parameditors.quickloader import QuickLoaderEditor
from utilitys import ParamEditorPlugin, ProcessIO, widgets as uw, ParamEditor, ParamContainer, ShortcutParameter
import s3a.shared
from s3a import models, XYVertices, ComplexXYVertices
from s3a.constants import PRJ_CONSTS as CNST, REQD_TBL_FIELDS as RTF, PRJ_ENUMS, GEN_PROPS_DIR, SCHEMES_DIR, \
  SHORTCUTS_DIR
from s3a.models import s3abase
from s3a.plugins.base import ProcessorPlugin


class MainImagePlugin(ParamEditorPlugin):
  name = __groupingName__ = 'Main Image'
  _makeMenuShortcuts = False

  def __init__(self):
    super().__init__()
    self.props = ParamContainer()

  def attachWinRef(self, win: s3abase.S3ABase):
    mainImg = win.mainImg

    copier = mainImg.regionCopier
    def startCopy():
      """
      Copies the selected components. They can be pasted by <b>double-clicking</b>
      on the destination location. When done copying, Click the *Clear ROI* tool change
      the current draw action.
      """
      copier.inCopyMode = True
      copier.sigCopyStarted.emit()
    def startMove():
      """
      Moves the selected components. They can be pasted by <b>double-clicking</b>
      on the destination location.
      """
      copier.inCopyMode = False
      copier.sigCopyStarted.emit()

    self.registerFunc(startMove, btnOpts=CNST.TOOL_MOVE_REGIONS)
    self.registerFunc(startCopy, btnOpts=CNST.TOOL_COPY_REGIONS)
    copier.sigCopyStopped.connect(win.mainImg.updateFocusedComp)


    disp = win.compDisplay
    # Wrap in process to ignore the default param

    def actHandler(verts, param):
      # When editing, only want to select if nothing is already started
      if (param not in [CNST.DRAW_ACT_REM, CNST.DRAW_ACT_ADD]
          or len(self.win.vertsPlg.region.regionData) == 0
      ):
        # Special case: Selection with point shape should be a point
        if self.win.mainImg.shapeCollection.curShapeParam == CNST.DRAW_SHAPE_POINT:
          verts = verts.mean(0, keepdims=True)
        # Make sure to check vertices plugin regions since they suppress disp's regions for focused ids
        # Must be done first, since a no-find in disp's regions will deselect them
        with uw.makeDummySignal(win.compTbl, 'sigSelectionChanged'):
          # Second call should handle the true selection signal
          disp.reflectSelectionBoundsMade(verts, self.win.vertsPlg.region)
          disp.reflectSelectionBoundsMade(verts, clearExisting=False)

        nonUniqueIds = win.compTbl.ids_rows_colsFromSelection(excludeNoEditCols=False, warnNoneSelection=False)[:,0]
        selection = pd.unique(nonUniqueIds)
        win.compTbl.sigSelectionChanged.emit(selection)

    win.sharedAttrs.generalProps.registerProp(CNST.PROP_MIN_COMP_SZ, container=self.props)
    self.registerFunc(disp.mergeSelectedComps, btnOpts=CNST.TOOL_MERGE_COMPS, ignoreKeys=['keepId'])
    self.registerFunc(disp.splitSelectedComps, btnOpts=CNST.TOOL_SPLIT_COMPS)
    win.mainImg.registerDrawAction([CNST.DRAW_ACT_ADD, CNST.DRAW_ACT_REM, CNST.DRAW_ACT_SELECT, CNST.DRAW_ACT_PAN],
                                 actHandler)
    self.registerFunc(disp.removeSelectedCompOverlap, btnOpts=CNST.TOOL_REM_OVERLAP)

    win.mainImg.registerDrawAction(CNST.DRAW_ACT_CREATE, self.createComponent)
    win.mainImg.addTools(self.toolsEditor)

    self.tableData = win.sharedAttrs.tableData

    # No need for a dropdown menu
    self.dock = None
    super().attachWinRef(win)

  @property
  def image(self):
    return self.win.mainImg.image

  def createComponent(self, roiVerts: XYVertices):
    verts = np.clip(roiVerts.astype(int), 0, self.image.shape[:2][::-1])

    if cv.contourArea(verts) < self.props[CNST.PROP_MIN_COMP_SZ]:
      # Use as selection instead of creation
      self.win.compDisplay.reflectSelectionBoundsMade(roiVerts[[0]])
      return

    # noinspection PyTypeChecker
    verts = ComplexXYVertices([verts])
    newComps = self.tableData.makeCompDf()
    newComps[RTF.VERTICES] = [verts]
    self.win.add_focusComps(newComps)

class CompTablePlugin(ParamEditorPlugin):
  name = 'Component Table'

  def __initEditorParams__(self, shared: s3a.shared.SharedAppSettings):
    super().__initEditorParams__()
    self.dock.addEditors([shared.filter])

  def attachWinRef(self, win: s3abase.S3ABase):

    tbl = win.compTbl
    for func, param in zip(
        [lambda: tbl.setSelectedCellsAs_gui(), tbl.removeSelectedRows_gui, tbl.setSelectedCellsAsFirst,
         lambda: win.compDisplay.scaleViewboxToSelectedIds()],
        [CNST.TOOL_TBL_SET_AS, CNST.TOOL_TBL_DEL_ROWS, CNST.TOOL_TBL_SET_SAME_AS_FIRST,
         CNST.TOOL_TBL_ZOOM_TO_COMPS]):
      param.opts['ownerObj'] = win.compTbl
      self.registerFunc(func, name=param.name, btnOpts=param)
    tbl.menu = self.toolsEditor.actionsMenuFromProcs(parent=tbl, nest=True)
    super().attachWinRef(win)
    self.tableData = win.sharedAttrs.tableData

class EditPlugin(ParamEditorPlugin):

  name = 'Edit'

  def attachWinRef(self, win: s3abase.S3ABase):
    super().attachWinRef(win)
    stack = win.sharedAttrs.actionStack

    self.registerFunc(stack.undo, name='Undo', btnOpts=CNST.TOOL_UNDO)
    self.registerFunc(stack.redo, name='Redo', btnOpts=CNST.TOOL_REDO)

    def updateUndoRedoTxts():
      self.undoAct.setText(f'Undo: {stack.undoDescr}')
      self.redoAct.setText(f'Redo: {stack.redoDescr}')
    stack.stackChangedCallbacks.append(updateUndoRedoTxts)
    updateUndoRedoTxts()

  @property
  def undoAct(self):
      return [a for a in self.menu.actions() if a.text().startswith('Undo')][0]

  @property
  def redoAct(self):
    return [a for a in self.menu.actions() if a.text().startswith('Redo')][0]

class RandomToolsPlugin(ParamEditorPlugin):
  name = 'Tools'
  _showFuncDetails = True

  def attachWinRef(self, win: s3abase.S3ABase):
    super().attachWinRef(win)

    self.registerFunc(self.showDevConsole_gui, name='Show Dev Console')
    self.registerFunc(win.clearBoundaries, btnOpts=CNST.TOOL_CLEAR_BOUNDARIES)
    self.registerFunc(win.compDisplay.exportCompOverlay, name='Export Component Overlay', toClipboard=True)
    self.registerFunc(lambda: win.setMainImg(None), name='Clear Current Image')

  def showDevConsole_gui(self):
    """
    Opens a console that allows dynamic interaction with current variables. If IPython
    is on your system, a qt console will be loaded. Otherwise, a (less capable) standard
    pyqtgraph console will be used.
    """
    namespace = dict(app=self.win, rtf=RTF)
    # "dict" default is to use repr instead of string for internal elements, so expanding
    # into string here ensures repr is not used
    nsPrintout = [f"{k}: {v}" for k, v in namespace.items()]
    text = f'Starting console with variables:\n' \
           f'{nsPrintout}'
    # Broad exception is fine, fallback is good enough. Too many edge cases to properly diagnose when Pycharm's event
    # loop is sync-able with the Jupyter dev console
    # noinspection PyBroadException
    try:
      console = uw.ConsoleWidget(parent=self.win, namespace=namespace, text=text)
    except Exception:
      # Ipy kernel can have issues for many different reasons. Always be ready to fall back to traditional console
      console = pg_console.ConsoleWidget(parent=self.win, namespace=namespace, text=text)
    console.setWindowFlags(QtCore.Qt.Window)
    console.show()

class HelpPlugin(ParamEditorPlugin):
  name = 'Help'

  def attachWinRef(self, win: QtWidgets.QMainWindow):
    super().attachWinRef(win)
    self.registerFunc(lambda: QtGui.QDesktopServices.openUrl(QtCore.QUrl('https://gitlab.com/ficsresearch/s3a/-/wikis/docs/user\'s-guide')),
                         name='Online User Guide')
    self.registerFunc(lambda: QtWidgets.QMessageBox.aboutQt(win, 'About Qt'), name='About Qt')
    self.registerFunc(self.iconAttributions_gui, name='Icon Attributions')

  def iconAttributions_gui(self):
    htmlStr = """
    <div>Icons made by <a href="https://www.freepik.com" title="Freepik">Freepik</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a></div>

    <div>Icons made by <a href="https://www.flaticon.com/authors/those-icons" title="Those Icons">Those Icons</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a></div>

    <div>Icons made by <a href="https://www.flaticon.com/authors/pixel-perfect" title="Pixel perfect">Pixel perfect</a> from <a href="https://www.flaticon.com/" title="Flaticon">www.flaticon.com</a></div>
    """

    QtWidgets.QMessageBox.information(self.win, 'Icon Attributions', htmlStr)


def miscFuncsPluginFactory(name_: str=None, regFuncs: Sequence[Callable]=None, titles: Sequence[str]=None, showFuncDetails=False):
  class FuncContainerPlugin(ParamEditorPlugin):
    name = name_
    _showFuncDetails = showFuncDetails

    def attachWinRef(self, win: s3abase.S3ABase):
      super().attachWinRef(win)

      nonlocal regFuncs, titles
      if regFuncs is None:
        regFuncs = []
      if titles is None:
        titles = [None] * len(regFuncs)
      for func, title in zip(regFuncs, titles):
        self.registerFunc(func, title)

  return FuncContainerPlugin


class MultiPredictionsPlugin(ProcessorPlugin):
  name = 'Multi-Predictions'

  mgr: models.tablemodel.ComponentMgr

  def __initEditorParams__(self, shared: s3a.shared.SharedAppSettings):
    super().__initEditorParams__()
    self.procEditor = shared.multiPredClctn.createProcessorEditor(type(self), self.name + ' Processor')
    self.dock.addEditors([self.procEditor])

  def __init__(self):
    super().__init__()
    self.registerFunc(self.lastRunAnalytics)

  def attachWinRef(self, win):
    super().attachWinRef(win)
    self.mgr = win.compMgr
    self.mainImg = win.mainImg
    win.mainImg.toolsEditor.registerFunc(self.predictFromSelection, btnOpts=CNST.TOOL_PRED_SEL)

  def makePrediction(self, comps: pd.DataFrame):
    if self.win.mainImg.image is None:
      return
    vbRange = np.array(self.mainImg.getViewBox().viewRange()).T
    newComps = self.curProcessor.run(components=comps, image=self.win.mainImg.image,
                                     viewbox=vbRange)
    if not isinstance(newComps, ProcessIO):
      newComps = ProcessIO(components=newComps)
    compsToAdd = newComps['components']
    addType = newComps.get('addType', PRJ_ENUMS.COMP_ADD_AS_NEW)
    self.mgr.addComps(compsToAdd, addType)


  def predictFromSelection(self):
    selectedIds = self.win.compDisplay.selectedIds
    if len(selectedIds) == 0:
      return
    self.makePrediction(self.mgr.compDf.loc[selectedIds])

  def lastRunAnalytics(self):
    raise NotImplementedError

class SettingsPlugin(ParamEditorPlugin):
  name = 'Settings'

  def __initEditorParams__(self, **kwargs):
    super().__initEditorParams__(**kwargs)

    self.generalProps = ParamEditor(saveDir=GEN_PROPS_DIR, fileType='genprops',
                                    name='App Settings')
    self.colorScheme = ParamEditor(saveDir=SCHEMES_DIR, fileType='scheme',
                                   name='Color Scheme')

    self.dock.addEditors([self.generalProps, self.colorScheme])


class ShortcutsPlugin(ParamEditorPlugin):
  name = 'Shortcuts'

  def __initEditorParams__(self, **kwargs):
    super().__initEditorParams__(**kwargs)

    self.shortcuts = ShortcutParameter.setRegistry(createIfNone=True, saveDir=SHORTCUTS_DIR)
    self.quickLoader = QuickLoaderEditor()

    self.dock.addEditors([self.shortcuts, self.quickLoader])




from ast import literal_eval
from datetime import datetime

import numpy as np

from ..structures import ComplexXYVertices, XYVertices, PrjParamGroup

from .componentio import ComponentIO, defaultIo
from .helpers import registerIoHandler

registerIoHandler(
  'ndarray',
  deserialize = lambda strVal: np.array(literal_eval(strVal)),
  serialize = lambda val: str(val.tolist()),
)

registerIoHandler(
  'complexxyvertices',
  deserialize = ComplexXYVertices.deserialize,
  serialize = ComplexXYVertices.serialize,
)

registerIoHandler(
  'xyvertices',
  deserialize = XYVertices.deserialize,
  serialize = XYVertices.serialize,
)

registerIoHandler(
  'bool',
  deserialize = lambda strVal: strVal.lower() == 'true',
)

registerIoHandler(
  'prjparam',
  deserialize = lambda param, strVal: PrjParamGroup.fieldFromParam(param.group, strVal),
  serialize = lambda param, val: str(val),
  takesParam=True
)

def listDeser(param, strVal):
  val = type(param.value)(strVal)
  fixedLims = param.opts.get('fixedLimits', False)
  lims = param.opts['limits']
  if fixedLims and val not in lims:
    raise ValueError(f'Limits cannot grow and "{val} is not in limits')
  elif val not in lims:
    lims.append(val)
  return val
registerIoHandler(
  'list',
  deserialize = listDeser,
  serialize = lambda param, val: str(val),
  takesParam=True
)

def superAnnTime(val):
  parsedTime = datetime.strptime(val, '%Y-%m-%dT%H:%M:%S.%fZ')
  return str(parsedTime)

registerIoHandler(
  'superanntime',
  deserialize = superAnnTime,
)


import pydoc
from functools import lru_cache
from s3a.shims import entry_points

def INTERNAL_PLUGINS():
  from .misc import MainImagePlugin, CompTablePlugin, RandomToolsPlugin, EditPlugin, HelpPlugin, MultiPredictionsPlugin
  from .file import FilePlugin
  from .tablefield import VerticesPlugin

  return [FilePlugin, EditPlugin, MultiPredictionsPlugin, VerticesPlugin, MainImagePlugin, CompTablePlugin,
          RandomToolsPlugin, HelpPlugin]

@lru_cache()
def EXTERNAL_PLUGINS():
  discoveredPlgs = entry_points().get('s3a.plugins', [])
  externPlgs = []
  # noinspection PyTypeChecker
  for ep in discoveredPlgs:
    # ALL_PLUGINS is usually exposed here
    member = pydoc.locate(ep.value)
    if member:
      plgs = member()
      for plg in plgs:
        if plg not in externPlgs:
          externPlgs.append(plg)
  return externPlgs


"""
All API discrepancies between different systems, python versions, etc. should be resolved here if remotely possible
"""
import sys

__all__ = ['entry_points', 'typing_extensions']

if sys.version_info < (3, 8):
    from importlib_metadata import entry_points
    import typing_extensions
else:
    from importlib.metadata import entry_points
    import typing as typing_extensions


from typing import Callable, Dict

import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtCore, QtGui
from skimage.draw import draw

from s3a.generalutils import orderContourPts, symbolFromVerts
from utilitys import PrjParam

from s3a.constants import PRJ_CONSTS
from s3a.structures import XYVertices, ComplexXYVertices

__all__ = ['RectROI', 'PlotDataROI', 'PolygonROI', 'PointROI', 'SHAPE_ROI_MAPPING']
from s3a.views.clickables import BoundScatterPlot

qe = QtCore.QEvent


class PlotDataROI(BoundScatterPlot):
  connected = True

  def __init__(self, endEvType=qe.MouseButtonRelease, addIntermedPts=True):
    super().__init__(pxMode=False, size=1)
    self.setBrush(pg.mkBrush(0, 0, 0, 100))
    self.setPen(pg.mkPen('w', width=0))
    self.vertices = XYVertices()

    self.endEvType = endEvType
    self.viableEvTypes = {
      qe.MouseButtonPress, qe.MouseMove, qe.MouseButtonRelease, qe.MouseButtonDblClick
    } - {endEvType}
    self.constructingRoi = False
    self.addIntermedPts = addIntermedPts
    self.firstPt = XYVertices()

  def addRoiPoints(self, pts: XYVertices):
    # noinspection PyTypeChecker
    pts = pts.astype(int)
    if self.vertices.size > 0 and np.all(pts == self.vertices[-1]):
      return
    if self.addIntermedPts or len(self.vertices) < 2:
      vertsToUse = np.vstack([self.vertices, pts])
    else:
      vertsToUse = np.vstack([self.firstPt, pts])
    refactored = self._refactorPoints(vertsToUse)
    verts = self.vertices = XYVertices(refactored)
    connectData = np.vstack([verts, verts[[0]]]).view(XYVertices)
    symbol, pos = symbolFromVerts(ComplexXYVertices([connectData]))
    self.setData(*pos.T, symbol=[symbol])


  def setRoiPoints(self, pts: XYVertices=None):
    if pts is None:
      pts = XYVertices()
    self.setData()
    self.vertices = XYVertices()
    self.firstPt = pts
    if pts.size > 0:
      self.addRoiPoints(pts)

  def _refactorPoints(self, vertices: np.ndarray):
    return vertices

  def updateShape(self, ev: QtGui.QMouseEvent, xyEvCoords: XYVertices):
    """
    See function signature for :func:`ExtendedROI.updateShape`
    """
    success = True
    verts = None
    constructingRoi = self.constructingRoi
    # If not left click, do nothing
    if (int(ev.buttons()) & QtCore.Qt.LeftButton) == 0 \
        and ev.button() != QtCore.Qt.LeftButton:
      return self.constructingRoi, verts
    evType = ev.type()
    if evType == qe.MouseButtonPress and not constructingRoi:
      # Need to start a new shape
      self.setRoiPoints(xyEvCoords)
      constructingRoi = True
    if evType in self.viableEvTypes:
      self.addRoiPoints(xyEvCoords)
      constructingRoi = True
    elif evType == self.endEvType:
      # Done drawing the ROI, complete shape, get vertices
      verts = self.vertices
      constructingRoi = False
    else:
      success = False
      # Unable to process this event

    if success:
      ev.accept()
    self.constructingRoi = constructingRoi
    return self.constructingRoi, verts

class RectROI(PlotDataROI):

  def __init__(self, endEvType=qe.MouseButtonRelease):
    super().__init__(endEvType)
    self.addIntermedPts = False

  def _refactorPoints(self, vertices: np.ndarray):
    square = XYVertices(np.vstack([vertices.min(0), vertices.max(0)]))
    return np.vstack([
      square[0],
      [square[0,0], square[1,1]],
      [square[1,0], square[1,1]],
      [square[1,0], square[0,1]]
    ])

class PolygonROI(PlotDataROI):

  def __init__(self):
    super().__init__(qe.MouseButtonDblClick, True)
    self.viableEvTypes.remove(qe.MouseMove)
    self.viableEvTypes.remove(qe.MouseButtonRelease)

class PointROI(PlotDataROI):
  roiRadius = 1

  @classmethod
  def updateRadius(cls, radius=1):
    cls.roiRadius = radius


  def __init__(self):
    super().__init__()
    self.setSize(3)
    self.constructTypes = {qe.MouseButtonPress, qe.MouseMove}

  def updateShape(self, ev: QtGui.QMouseEvent, xyEvCoords: XYVertices):
    success = False
    verts = None
    constructingRoi = False
    if ev.type() in self.constructTypes:
      success = True
      self.setData(*xyEvCoords.T, size=self.roiRadius*2)
      verts = XYVertices(np.column_stack(draw.disk(xyEvCoords[0], self.roiRadius)))
      self.vertices = verts
      constructingRoi = True

    if success:
      ev.accept()
    return constructingRoi, verts

class EllipseROI(PlotDataROI):

  def __init__(self):
    super().__init__(endEvType=qe.MouseButtonRelease, addIntermedPts=False)

  def _refactorPoints(self, vertices: np.ndarray):
    pts = vertices[:, ::-1].astype(float)
    ma = pts.max(0)
    mi = pts.min(0)
    center = (ma - mi)/2 + mi
    normedR = ma[0]-mi[0]
    normedC = ma[1]-mi[1]
    # Apply scaling so mouse point is on perimeter
    # angle = np.arctan(normedR/normedC)
    # normedR = abs(np.cos(angle))*normedR
    # normedC = abs(np.sin(angle))*normedC
    sqr2 = np.sqrt(2)
    perim = draw.ellipse_perimeter(*center.astype(int), int(normedR/sqr2), int(normedC/sqr2))
    # Reorder to ensure no criss-crossing when these vertices are plotted
    perim = orderContourPts(np.column_stack(perim[::-1]))
    return perim.view(XYVertices)

SHAPE_ROI_MAPPING: Dict[PrjParam, Callable[[], PlotDataROI]] = {
  PRJ_CONSTS.DRAW_SHAPE_RECT: RectROI,
  PRJ_CONSTS.DRAW_SHAPE_FREE: PlotDataROI,
  PRJ_CONSTS.DRAW_SHAPE_POLY: PolygonROI,
  PRJ_CONSTS.DRAW_SHAPE_ELLIPSE: EllipseROI,
  PRJ_CONSTS.DRAW_SHAPE_POINT: PointROI,
}



from __future__ import annotations

import copy
from collections import deque, namedtuple, defaultdict
import typing as t
from functools import lru_cache
from warnings import warn

import numpy as np
import pandas as pd
import pyqtgraph as pg

import s3a
from s3a import PRJ_CONSTS as CNST, XYVertices, REQD_TBL_FIELDS as RTF, ComplexXYVertices
from s3a.processing.algorithms.imageproc import procCache
from s3a.structures import BlackWhiteImg
from s3a.views.regions import MultiRegionPlot, makeMultiRegionDf
from utilitys import PrjParam, DeferredActionStackMixin as DASM, ProcessStage, ProcessIO, RunOpts, fns
from .base import TableFieldPlugin
from ..constants import PRJ_ENUMS
from ..generalutils import getCroppedImg, showMaskDiff, tryCvResize, incrStageNames
from ..graphicsutils import RegionHistoryViewer
from ..shared import SharedAppSettings


class _REG_ACCEPTED: pass
buffEntry = namedtuple('buffentry', 'id_ vertices')

class VerticesPlugin(DASM, TableFieldPlugin):
  name = 'Vertices'

  def __initEditorParams__(self, shared: SharedAppSettings):
    super().__initEditorParams__()
    self.procEditor = shared.imgProcClctn.createProcessorEditor(type(self), self.name + ' Processor')

    self.dock.addEditors([self.procEditor])

  def __init__(self):
    super().__init__()
    self.region = MultiRegionPlot()
    self.region.hide()
    self.firstRun = True
    self.playbackWindow = RegionHistoryViewer()
    self.regionBuffer = deque(maxlen=CNST.PROP_UNDO_BUF_SZ.value)
    self.stageInfoImage = pg.ImageItem()
    self.stageInfoImage.hide()
    self._displayedStage = ''

    _, self._overlayParam = self.procEditor.registerFunc(self.overlayStageInfo, parentParam=self.procEditor._metaParamGrp,
                                            returnParam=True, runOpts=RunOpts.ON_CHANGED)

  def attachWinRef(self, win: s3a.S3A):
    win.mainImg.addItem(self.region)
    win.mainImg.addItem(self.stageInfoImage)

    def resetRegBuff(_, newSize):
      newBuff = deque(maxlen=newSize)
      newBuff.extend(self.regionBuffer)
      self.regionBuffer = newBuff
    mainBufSize = win.sharedAttrs.generalProps.params.child(win.__groupingName__, 'maxLength')
    mainBufSize.sigValueChanged.connect(resetRegBuff)

    funcLst = [self.resetFocusedRegion, self.fillRegionMask, self.clearFocusedRegion, self.clearProcessorHistory]
    paramLst = [CNST.TOOL_RESET_FOC_REGION, CNST.TOOL_FILL_FOC_REGION,
                CNST.TOOL_CLEAR_FOC_REGION, CNST.TOOL_CLEAR_HISTORY]
    for func, param in zip(funcLst, paramLst):
      self.registerFunc(func, btnOpts=param)

    def onChange():
      self.firstRun = True
      self.clearFocusedRegion()
      self.stageInfoImage.hide()
    win.mainImg.imgItem.sigImageChanged.connect(onChange)

    win.mainImg.registerDrawAction([CNST.DRAW_ACT_ADD, CNST.DRAW_ACT_REM], self._run_drawAct)
    win.mainImg.addTools(self.toolsEditor)
    self.vb: pg.ViewBox = win.mainImg.getViewBox()
    super().attachWinRef(win)

  def fillRegionMask(self):
    """Completely fill the focused region mask"""
    if self.mainImg.compSer is None: return
    filledImg = np.ones(self.mainImg.image.shape[:2], dtype='uint16')
    self.updateRegionFromMask(filledImg)

  @classmethod
  def clearProcessorHistory(cls):
    """
    Each time an update is made in the processor, it is saved so algorithmscan take
    past edits into account when performing their operations. Clearing that history
    will erase algorithm knowledge of past edits.
    """
    procCache['mask'].fill(0)

  def updateFocusedComp(self, newComp:pd.Series = None):
    self.updateFocusedPen_Fill()
    if self.mainImg.compSer[RTF.INST_ID] == -1:
      self.updateRegionFromDf(None)
      return
    oldId = self.mainImg.compSer[RTF.INST_ID]
    self.updateRegionFromDf(self.mainImg.compSer_asFrame)
    if newComp is None or oldId != newComp[RTF.INST_ID]:
      self.firstRun = True

  def updateFocusedPen_Fill(self):
    disp = self.win.compDisplay
    plt: MultiRegionPlot = disp.regionPlot
    mgrDf = self.win.compMgr.compDf
    regDf = plt.regionData.copy()
    focused = regDf[PRJ_ENUMS.FIELD_FOCUSED].to_numpy(bool)
    # Show IDs that were previously hidden
    for showId in regDf.index[regDf[RTF.VERTICES].apply(ComplexXYVertices.isEmpty)]:
      regDf.at[showId, RTF.VERTICES] = mgrDf.at[showId, RTF.VERTICES]

    # Make 0-area regions where the plot must be hidden
    for hideId in regDf.index[focused]:
      # Bug: Can't set array of values without indexer error since pandas thinks ComplexXYVertices is a 2d array
      regDf.at[hideId, RTF.VERTICES] = ComplexXYVertices()
    plt.resetRegionList(regDf, self.win.compDisplay.labelCol)

  def _run_drawAct(self, verts: XYVertices, param: PrjParam):
    # noinspection PyTypeChecker
    verts : XYVertices = verts.astype(int)
    if param == CNST.DRAW_ACT_ADD:
      self.run(fgVerts=verts)
    else:
      self.run(bgVerts=verts)

  def run(self, fgVerts: XYVertices=None, bgVerts: XYVertices=None):
    vertsDict = {}
    if fgVerts is not None:
      vertsDict['fgVerts'] = fgVerts
    if bgVerts is not None:
      vertsDict['bgVerts'] = bgVerts
    img = self.mainImg.image
    if img is None:
      compGrayscale = None
      compMask = None
    else:
      compGrayscale = self.region.toGrayImg(img.shape[:2])
      compMask = compGrayscale > 0
    # TODO: When multiple classes can be represented within focused image, this is where
    #  change will have to occur
    # Clip viewrange to min view axis area instead of max, which will happen internally
    # otherwise
    viewbox = self.mainImg.viewboxCoords()
    newGrayscale = self.curProcessor.run(
      image=img,
      prevCompMask=compMask,
      **vertsDict,
      firstRun=self.firstRun,
      viewbox=XYVertices(viewbox),
      prevCompVerts=ComplexXYVertices([r.stack() for r in self.region.regionData[RTF.VERTICES]])
    )
    if isinstance(newGrayscale, dict):
      newGrayscale = newGrayscale['image']
    newGrayscale = newGrayscale.astype('uint8')

    matchNames = incrStageNames(self.curProcessor.processor.stages_flattened)
    type(self).displayableInfos.fget.cache_clear()
    if self._displayedStage in matchNames:
      self.overlayStageInfo(self._displayedStage, self.stageInfoImage.opacity())
    else:
      self.stageInfoImage.hide()
    # Can't set limits to actual infos since equality comparison fails in pyqtgraph setLimits
    limits = [''] + list(self.displayableInfos)
    self._overlayParam.child('info').setLimits(limits)

    self.firstRun = False
    if not np.array_equal(newGrayscale, compGrayscale):
      self.updateRegionFromMask(newGrayscale)

  def updateRegionFromDf(self, newData: pd.DataFrame=None, offset: XYVertices=None):
    """
    Updates the current focused region using the new provided vertices
    :param newData: Dataframe to use.If *None*, the image will be totally reset and the component
      will be removed. Otherwise, the provided value will be used. For column information,
      see `makeMultiRegionDf`
    :param offset: Offset of newVerts relative to main image coordinates
    """
    fImg = self.mainImg
    # Since some calls to this function make an undo entry and others don't, be sure to save state for the undoable ones
    # revertId is only used when changedComp is true and an undo is valid
    newId = fImg.compSer[RTF.INST_ID]
    if newData is None or np.all(newData[RTF.VERTICES].apply(ComplexXYVertices.isEmpty)):
      newData = makeMultiRegionDf(0)

    self._maybeChangeFocusedComp(newData.index)

    if fImg.image is None:
      self.region.clear()
      self.regionBuffer.append(buffEntry(newId, ComplexXYVertices()))
      return

    oldData = self.region.regionData

    if offset is None:
      offset = XYVertices([[0,0]])
    # 0-center new vertices relative to FocusedImage image
    centeredData = newData
    if np.any(offset != 0):
      # Make a deep copy of verts (since they change) to preserve redos
      centeredData = centeredData.copy()
      centeredData[RTF.VERTICES] = self.applyOffset([copy.deepcopy(v) for v in newData[RTF.VERTICES]], offset)
    if oldData.equals(centeredData):
      return

    lblCol = self.win.compDisplay.labelCol
    self.region.resetRegionList(newRegionDf=centeredData, lblField=lblCol)
    self.region.focusById(centeredData.index)

    buffVerts = ComplexXYVertices()
    for inner in centeredData[RTF.VERTICES]: buffVerts.extend(inner)
    self.regionBuffer.append(buffEntry(newId, buffVerts))

  @staticmethod
  def applyOffset(complexVerts: t.Sequence[ComplexXYVertices], offset: XYVertices):
    centeredVerts = []
    for complexVerts in complexVerts:
      newVertList = ComplexXYVertices()
      for vertList in complexVerts:
        newVertList.append(vertList + offset)
      centeredVerts.append(newVertList)
    return centeredVerts

  def _maybeChangeFocusedComp(self, newIds: t.Sequence[int]):
    regionId = newIds[0] if len(newIds) else -1
    focusedId = self.mainImg.compSer[RTF.INST_ID]
    updated = regionId != focusedId
    if updated:
      if regionId in self.win.compMgr.compDf.index:
        self.win.compDisplay.selectRowsById([regionId])
      else:
        self.win.compDisplay.selectRowsById([])
    return updated

  def updateRegionFromMask(self, mask: BlackWhiteImg, offset=None, compId=None):
    if offset is None:
      offset = XYVertices([0,0])
    data = self.region.regionData.copy()
    if compId is None:
      compId = data.index[0] if len(data) else -1
    df = makeMultiRegionDf(vertices=[ComplexXYVertices.fromBwMask(mask)], idList=[compId])
    self.updateRegion_undoable(df, offset=offset)


  @DASM.undoable('Modify Focused Component')
  def updateRegion_undoable(self, newData: pd.DataFrame=None, offset: XYVertices=None):
    oldData = self.region.regionData.copy()
    self.updateRegionFromDf(newData, offset=offset)
    yield
    self.updateRegionFromDf(oldData)

  def acceptChanges(self, overrideVerts: ComplexXYVertices=None):
    # Add in offset from main image to VertexRegion vertices
    newVerts = overrideVerts or self.collapseRegionVerts()
    ser = self.mainImg.compSer
    ser[RTF.VERTICES] = newVerts
    self.updateFocusedComp()

  def collapseRegionVerts(self, simplify=True):
    """
    Region can consist of multiple separate complex vertices. However, the focused series
    can only contain one list of complex vertices. This function collapses all data in self.region
    into one list of complex vertices.

    :param simplify: Overlapping regions can be simplified by converting to and back from
      an image. This can be computationally intensive at times, in which case `simplify` can
      be set to *False*
    """
    outVerts = ComplexXYVertices([verts for cplxVerts in self.region.regionData[RTF.VERTICES] for verts in cplxVerts])
    if simplify:
      outVerts = ComplexXYVertices.fromBwMask(outVerts.toMask())
    return outVerts

  def clearFocusedRegion(self):
    # Reset drawn comp vertices to nothing
    # Only perform action if image currently exists
    if self.mainImg.compSer is None:
      return
    self.updateRegion_undoable(None)

  def resetFocusedRegion(self):
    """Reset the focused image by restoring the region mask to the last saved state"""
    if self.mainImg.compSer is None:
      return
    self.updateRegion_undoable(self.mainImg.compSer_asFrame)

  def _onActivate(self):
    self.region.show()
    self.updateFocusedPen_Fill()

  def _onDeactivate(self):
    self.region.hide()
    self.win.compDisplay.regionPlot.updateColors()

  def getRegionHistory(self):
    outImgs = []
    if not self.regionBuffer:
      return None, []
    firstId = self.regionBuffer[-1].id_
    bufferRegions = [buf.vertices for buf in self.regionBuffer if buf.id_ == firstId]

    if not bufferRegions:
      return None, []

    # First find offset and img size so we don't
    # have to keep copying a full image sized output every time
    allVerts = np.vstack([v.stack() for v in bufferRegions])
    initialImg, slices = getCroppedImg(self.mainImg.image, allVerts)
    imShape = initialImg.shape[:2]
    offset = slices[0]
    img = np.zeros(imShape, bool)
    outImgs.append(img)
    for singleRegionVerts in bufferRegions:
      # Copy to avoid screwing up undo buffer!
      copied = singleRegionVerts.removeOffset(offset)
      img = copied.toMask(imShape, warnIfTooSmall=False)
      outImgs.append(img)
    return initialImg, outImgs

  def overlayStageInfo(self, info: t.Union[str, dict]='', alpha=1.0):
    """
    :param info:
      pType: list
      helpText: "If a name is given, this stage's info result will be shown on top of the image. Note that
        if multiple stages exist with the same name and a string is passed, include the 1-based numeric index in the
        name, i.e. if two 'Open' stages exist, to select the second stage pass 'Open#2'.
      "
      limits: ['']
    :param alpha:
      helpText: Opacity of this overlay
      limits: [0, 1]
      step: 0.1
    """
    if not isinstance(info, dict):
      self._displayedStage = info
      info = self.displayableInfos.get(info, None)
    else:
      self._displayedStage = info['name']
    if not info:
      # Reset to none
      self.stageInfoImage.hide()
      return
    useImg = info['image']
    span = info['span']
    pos = info['pos']
    if useImg.shape[:2] != span:
      useImg = tryCvResize(useImg, span[::-1], asRatio=False)
    self.stageInfoImage.setPos(*pos)
    self.stageInfoImage.setImage(useImg)
    self.stageInfoImage.setOpts(opacity=alpha)
    self.stageInfoImage.show()

  @property
  @lru_cache()
  def displayableInfos(self):
    outInfos = {}
    stages = self.curProcessor.processor.stages_flattened
    matchNames = incrStageNames(stages)
    boundSlices = None

    for name, stage in zip(matchNames, stages):
      if stage.result is None:
        # Not run yet
        continue
      info = self.curProcessor.processor.singleStageInfo(stage)
      if info:
        # TODO: Figure out which 'info' to use
        info = info[0]
      if info is None or 'image' not in info:
        continue
      useImg = info['image']
      # This image was likely from a cropped area, make sure to put it in the right place
      imPos = (0, 0)
      span = useImg.shape[:2]
      if boundSlices is not None:
        # Resizing may have also occurred
        span = []
        imPos = []
        for curSlice in boundSlices: # type: slice
          # Pos is specified in x-y, not row-col
          curSlice: slice
          imPos.append(curSlice.start)
          span.append(curSlice.stop - curSlice.start)
        span = tuple(span)
        imPos = tuple(imPos)
      outInfos[name] = {
        'name': name,
        'image': useImg,
        'span': span,
        'pos': imPos[::-1]
      }
      # Bound slices up to the evaluated stage are used for resizing information
      if isinstance(stage.result, ProcessIO) and 'boundSlices' in stage.result:
        boundSlices = stage.result['boundSlices']
    return outInfos


  def playbackRegionHistory(self):
    initialImg, history = self.getRegionHistory()
    if initialImg is None:
      warn('No edits found, nothing to do', UserWarning)
      return
    # Add current state as final result
    history += [history[-1]]
    diffs = [showMaskDiff(o, n) for (o, n) in zip(history, history[1:])]
    self.playbackWindow.setDiffs(diffs)
    self.playbackWindow.displayPlt.setImage(initialImg)
    self.playbackWindow.show()
    self.playbackWindow.raise_()


from __future__ import annotations

import os
import sys

import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets

__all__ = ['appInst', 'S3A', 'REQD_TBL_FIELDS',
           'ComplexXYVertices', 'XYVertices', 'PRJ_CONSTS', 'ComponentIO',
           'ProjectData', '__version__', 'PRJ_ENUMS']

pg.setConfigOptions(imageAxisOrder='row-major')

from .__version__ import __version__

# Makes sure that when the folder is run as a module, the app exists in the outermost
# scope of the application
customPlatform = os.environ.get('S3A_PLATFORM')
args = list(sys.argv)
if customPlatform is not None:
  args += ['-platform', customPlatform]
sys.argv = args
appInst = pg.mkQApp()
# Allow selectable text in message boxes
appInst.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")

from . import graphicsutils as gutils


# Import here to resolve resolution order
import s3a.constants
import s3a.structures

from ._io import ComponentIO, defaultIo
from s3a.structures import XYVertices, ComplexXYVertices
from s3a.constants import REQD_TBL_FIELDS, PRJ_CONSTS, CFG_DIR, PRJ_ENUMS

from .views.s3agui import S3A

from s3a.plugins.misc import RandomToolsPlugin
from s3a.plugins.file import ProjectData



from __future__ import annotations

import traceback
import warnings

import numpy as np

from s3a.generalutils import augmentException
from s3a.structures import ComplexXYVertices, XYVertices
from utilitys import NestedProcWrapper
from utilitys.fns import warnLater
from .processing import *

__all__ = ['ImgProcWrapper', 'NestedProcWrapper']

class ImgProcWrapper(NestedProcWrapper):
  output: np.ndarray

  def run(self, **kwargs):
    newIo = self._ioDictFromRunKwargs(kwargs)
    try:
      with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        result = self.processor.run(newIo)
    except Exception as ex:
      msg = 'Exception during processor run:\n'
      augmentException(ex, msg)
      result = ProcessIO(image=kwargs['prevCompMask'])
      warnLater(str(ex), UserWarning)

    outImg = result['image'].astype(bool)
    if outImg.ndim > 2:
      outImg = np.bitwise_or.reduce(outImg, 2)
    self.output = outImg
    return outImg

  def resultAsVerts(self, localEstimate=True):
    initialList = ComplexXYVertices.fromBwMask(self.output)
    if len(initialList) == 0:
      return initialList
    if not localEstimate:
      return [ComplexXYVertices([lst]) for lst in initialList]
    # else, all vertices belong to the same component
    else:
      return [initialList]

  @staticmethod
  def _ioDictFromRunKwargs(runKwargs):
    image = runKwargs.get('image', None)
    if image is None:
      raise RuntimeError('Cannot run processor without an image')

    runKwargs.setdefault('firstRun', True)
    for name in 'fgVerts', 'bgVerts':
      runKwargs.setdefault(name, XYVertices())

    if runKwargs.get('prevCompMask', None) is None:
      noPrevMask = True
      runKwargs['prevCompMask'] = np.zeros(image.shape[:2], bool)
    else:
      noPrevMask = False

    return ProcessIO(**runKwargs, noPrevMask=noPrevMask)


import typing as t
from typing import Tuple, Union, Dict, Any

import cv2 as cv
import numpy as np
from scipy.ndimage import binary_fill_holes
from skimage import morphology as morph, img_as_float
from skimage import segmentation as seg
from skimage.measure import regionprops, label
from skimage.morphology import flood

from s3a.generalutils import cornersToFullBoundary, getCroppedImg, imgCornerVertices, \
  showMaskDiff, tryCvResize
from s3a.structures import BlackWhiteImg, XYVertices, ComplexXYVertices, NChanImg, GrayImg
from utilitys import fns
from utilitys.processing import *

# TODO: Establish better mechanism than global buffer
procCache: Dict[str, Any] = {
  'mask': np.array([[]], 'uint8')
}
"""
While a global structure is not ideal, it allows algorithms to use previous results across multiple calls.
Only results that need to be accessed across several different functions should be stored here. If one function
needs to maintain its state across multiple calls, it should be promoted to an AtomicProcess and use its own
`result` structure.
keys are:
  mask: Mask of previous results, where 0 = unspecified, 1 = background, 2 = foreground
"""

def growSeedpoint(img: NChanImg, seeds: XYVertices, thresh: float) -> BlackWhiteImg:
  shape = np.array(img.shape[0:2])
  bwOut = np.zeros(shape, dtype=bool)
  # Turn x-y vertices into row-col seeds
  seeds = seeds[:, ::-1]
  # Remove seeds that don't fit in the image
  seeds = seeds[np.all(seeds >= 0, 1)]
  seeds = seeds[np.all(seeds < shape, 1)]

  for seed in seeds:
    for chan in range(img.shape[2]):
      curBwMask = flood(img[...,chan], tuple(seed), tolerance=thresh)
      bwOut |= curBwMask
  return bwOut

# def _growSeedpoint_cv_fastButErratic(img: NChanImg, seeds: XYVertices, thresh: float):
#   if len(seeds) == 0:
#     return np.zeros(img.shape[:2], bool)
#   nChans = img.shape[2] if img.ndim > 2 else 1
#   thresh = int(np.clip(thresh, 0, 255))
#   imRCShape = np.array(img.shape[:2])
#   bwOut = np.zeros(imRCShape+2, 'uint8')
#   # Throw away seeds outside image boundaries
#   seeds = seeds[np.all(seeds < imRCShape, axis=1)]
#   seeds = np.fliplr(seeds)
#   mask = np.zeros(imRCShape+2, 'uint8')
#   for seed in seeds:
#     mask.fill(0)
#     flooded = img.copy()
#     seed = tuple(seed.flatten())
#     _, _, curOut, _ = cv.floodFill(flooded, mask, seed, 255, (thresh,)*nChans, (thresh,)*nChans,8)
#     bwOut |= curOut
#   bwOut = bwOut[1:-1,1:-1]
#   return bwOut.astype(bool)

def _cvConnComps(image: np.ndarray, returnLabels=True, areaOnly=True, removeBg=True):
  if image.dtype != 'uint8':
    image = image.astype('uint8')
  _, labels, conncomps, _ = cv.connectedComponentsWithStats(image)
  startIdx = 1 if removeBg else 0
  if areaOnly:
    conncomps = conncomps[:, cv.CC_STAT_AREA]
  if returnLabels:
    return conncomps[startIdx:], labels
  return conncomps[startIdx:]

def format_vertices(image: NChanImg, fgVerts: XYVertices, bgVerts: XYVertices,
                    prevCompMask: BlackWhiteImg, firstRun: bool,
                    useFullBoundary=True,
                    keepVertHistory=True):

  if firstRun or not keepVertHistory:
    _historyMask = np.zeros(image.shape[:2], 'uint8')
  else:
    _historyMask = procCache['mask']

  asForeground = True
  # 0 = unspecified, 1 = background, 2 = foreground
  for fillClr, verts in enumerate([bgVerts, fgVerts], 1):
    if not verts.empty and verts.connected:
      cv.fillPoly(_historyMask, [verts], fillClr)

  if useFullBoundary:
    if not fgVerts.empty:
      fgVerts = cornersToFullBoundary(fgVerts)
    if not bgVerts.empty:
      bgVerts = cornersToFullBoundary(bgVerts)

  if fgVerts.empty and bgVerts.empty:
    # Give whole image as input, trim from edges
    fgVerts = imgCornerVertices(image)
    fgVerts = cornersToFullBoundary(fgVerts)
    _historyMask[fgVerts.rows, fgVerts.cols] = 1
  procCache['mask'] = _historyMask
  curHistory = _historyMask.copy()
  if fgVerts.empty:
    # Invert the mask and paint foreground pixels
    asForeground = False
    # Invert the history mask too
    curHistory[_historyMask == 2] = 1
    curHistory[_historyMask == 1] = 2
    fgVerts = bgVerts
    bgVerts = XYVertices()

  if asForeground:
    foregroundAdjustedCompMask = prevCompMask.copy()
  else:
    foregroundAdjustedCompMask = ~prevCompMask

  # Default to bound slices that encompass the whole image
  bounds = np.array([[0, 0], image.shape[:2][::-1]])
  boundSlices = slice(*bounds[:,1]), slice(*bounds[:,0])
  return ProcessIO(image=image, summaryInfo=None, fgVerts=fgVerts, bgVerts=bgVerts, asForeground=asForeground,
                   historyMask=curHistory, prevCompMask=foregroundAdjustedCompMask,
                   origCompMask=prevCompMask, boundSlices=boundSlices)

def crop_to_local_area(image: NChanImg,
                       fgVerts: XYVertices,
                       bgVerts: XYVertices,
                       prevCompMask: BlackWhiteImg,
                       prevCompVerts: ComplexXYVertices,
                       viewbox: XYVertices,
                       historyMask: GrayImg,
                       reference='viewbox',
                       margin_pct=10,
                       maxSize=0,
                       useMinSpan=False
                       ):
  """
  :param reference:
    pType: list
    limits:
      - image
      - component
      - viewbox
      - roi
  :param maxSize: Maximum side length for a local portion of the image. If the local area exceeds this, it will be
    rescaled to match this size. It can be beneficial for algorithms that take a long time to run, and quality of
    segmentation can be retained. Set to <= 0 to have no maximum size
  :param useMinSpan: When `viewbox` is the reference, this determines whether to crop to the area defined by the
    shortest side of the viewbox. So, for aspect ratios far from 1 (i.e. heavily rectangular), this prevents a large
    area from being used every time
  """
  roiVerts = np.vstack([fgVerts, bgVerts])
  compVerts = np.vstack([prevCompVerts.stack(), roiVerts])
  if reference == 'image':
    allVerts = np.array([[0, 0], image.shape[:2]])
  elif reference == 'roi' and len(roiVerts) > 1:
    allVerts = roiVerts
  elif reference == 'component' and len(compVerts) > 1:
    allVerts = compVerts
  else:
    # viewbox or badly sized previous region/roi
    if useMinSpan:
      center = viewbox.mean(0)
      spans = center - viewbox[0]
      adjustments = (spans - min(spans)).astype(viewbox.dtype)
      adjustments = (spans - min(spans)).astype(viewbox.dtype)
      # maxs need to be subtracted toward center, mins need to be extended toward center
      dim = np.argmax(adjustments)
      adjust = adjustments[dim]
      maxs = viewbox[:, dim] == viewbox[:, dim].max()
      mins = ~maxs
      viewbox[maxs, dim] -= adjust
      viewbox[mins, dim] += adjust
    allVerts = np.vstack([viewbox, roiVerts])
  # Lots of points, use their bounded area
  try:
    vertArea_rowCol = (allVerts.max(0)-allVerts.min(0))[::-1]
  except ValueError:
    # 0-sized
    vertArea_rowCol = 0
  margin = int(round(max(vertArea_rowCol) * (margin_pct / 100)))
  cropped, bounds = getCroppedImg(image, allVerts, margin)
  ratio = 1
  curMaxDim = np.max(cropped.shape[:2])
  if 0 < maxSize < curMaxDim:
    ratio = maxSize / curMaxDim

  vertOffset = bounds.min(0)
  useVerts = [fgVerts, bgVerts]
  for ii in range(2):
    # Add additional offset
    tmp = (((useVerts[ii] - vertOffset)*ratio)).astype(int)
    useVerts[ii] = np.clip(tmp, a_min=[0,0], a_max=(bounds[1,:]-1)*ratio, dtype=int, casting='unsafe')
  fgVerts, bgVerts = useVerts

  boundSlices = slice(*bounds[:,1]), slice(*bounds[:,0])
  croppedCompMask = prevCompMask[boundSlices]
  curHistory = historyMask[boundSlices]

  rectThickness = int(max(1, *image.shape)*0.005)
  toPlot = image.copy()
  borderMask = cv.rectangle(np.zeros(image.shape[:2], dtype='uint8'), tuple(bounds[0,:]), tuple(bounds[1,:]),
                        255, rectThickness) > 0
  if image.ndim < 3:
    borderClr = 255
    borderSlice = borderMask,
  else:
    borderClr = [255] + [0 for _ in range(image.shape[2]-1)]
    borderSlice = borderMask, slice(None, None)
  borderClr = np.clip(np.array(borderClr, dtype=image.dtype), image.min(), image.max(), dtype=image.dtype)
  toPlot[borderSlice] = borderClr
  info = {'name': 'Selected Area', 'image': toPlot}
  out = ProcessIO(image=cropped, fgVerts=fgVerts, bgVerts=bgVerts, prevCompMask=croppedCompMask,
                   boundSlices=boundSlices, historyMask=curHistory, resizeRatio=ratio, summaryInfo=info)
  if ratio < 1:
    for kk in 'image', 'prevCompMask', 'historyMask':
      out[kk] = cv_resize(out[kk], ratio, interpolation='INTER_NEAREST')
  return out

def apply_process_result(image: NChanImg, asForeground: bool,
                         prevCompMask: BlackWhiteImg, origCompMask: BlackWhiteImg,
                         boundSlices: Tuple[slice,slice],
                         resizeRatio: float):
  if asForeground:
    bitOperation = np.bitwise_or
  else:
    # Add to background
    bitOperation = lambda curRegion, other: ~(curRegion | other)
  # The other basic operations need the rest of the component mask to work properly,
  # so expand the current area of interest only as much as needed. Returning to full size
  # now would incur unnecessary addtional processing times for the full-sized image
  outMask = origCompMask.copy()
  change = bitOperation(prevCompMask, image)
  if resizeRatio < 1:
    origSize = (boundSlices[0].stop - boundSlices[0].start,
                boundSlices[1].stop - boundSlices[1].start)
    # Without first converting to float, interpolation will be cliped to True/False. This causes
    # 'jagged' edges in the output
    change = cv_resize(change.astype(float), origSize[::-1], asRatio=False, interpolation='INTER_LINEAR')
  # Vast majority of samples will be near 0 or 1, filter these out
  subset = change[(change > 0.01) & (change < 0.99)].ravel()
  if len(subset) == 0:
    change = change > 0
  else:
    change = change > subset.mean()
  outMask[boundSlices] = change
  foregroundPixs = np.c_[np.nonzero(outMask)]
  # Keep algorithm from failing when no foreground pixels exist
  if len(foregroundPixs) == 0:
    mins = [0,0]
    maxs = [1,1]
  else:
    mins = foregroundPixs.min(0)
    maxs = foregroundPixs.max(0)

  # Add 1 to max slice so stopping value is last foreground pixel
  newSlices = (slice(mins[0], maxs[0]+1), slice(mins[1], maxs[1]+1))
  return ProcessIO(image=outMask[newSlices], boundSlices=newSlices, summaryInfo={'image': change})

def return_to_full_size(image: NChanImg, origCompMask: BlackWhiteImg,
                        boundSlices: Tuple[slice]):
  out = np.zeros_like(origCompMask)
  if image.ndim > 2:
    image = image.mean(2).astype(int)
  out[boundSlices] = image

  infoMask = showMaskDiff(origCompMask[boundSlices], image)

  return ProcessIO(image=out, summaryInfo={'image': infoMask, 'name': 'Finalize Region'})

def fill_holes(image: NChanImg):
  return ProcessIO(image=binary_fill_holes(image))

def disallow_paint_tool(_image: NChanImg, fgVerts: XYVertices, bgVerts: XYVertices):
  if len(np.vstack([fgVerts, bgVerts])) < 2:
    raise ValueError('This algorithm requires an enclosed area to work.'
                              ' Only one vertex was given as an input.')
  return ProcessIO(image=_image)


@fns.dynamicDocstring(morphOps=[d for d in dir(cv) if d.startswith('MORPH_')])
def morph_op(image: NChanImg, radius=1, op: str='', shape='rectangle'):
  """
  :param radius: Radius of the structuring element. Note that the total side length
    of the structuring element will be (2*radius)+1.
  :param shape:
    helpText: Structuring element shape
    pType: list
    limits:
      - rectangle
      - disk
      - diamond
  :param op:
    pType: list
    limits: {morphOps}
  """
  opType = getattr(cv, op)
  if image.ndim > 2:
    image = image.mean(2)
  image = image.astype('uint8')
  ksize = [radius]
  if shape == 'rectangle':
    ksize = [ksize[0]*2+1]*2
  strel = getattr(morph, shape)(*ksize)
  outImg = cv.morphologyEx(image.copy(), opType, strel)
  return ProcessIO(image=outImg)

def opening_factory():
  return AtomicProcess(morph_op, 'Opening', op='MORPH_OPEN')

def closing_factory():
  return AtomicProcess(morph_op, 'Closing', op='MORPH_CLOSE')

def keep_largest_comp(image: NChanImg):
  if not np.any(image):
    return ProcessIO(image=image)
  areas, labels = _cvConnComps(image)
  # 0 is background, so skip it
  out = np.zeros_like(image, shape=image.shape[:2])
  # Offset by 1 since 0 was removed earlier
  maxAreaIdx = np.argmax(areas) + 1
  out[labels == maxAreaIdx] = True
  return ProcessIO(image=out)

def remove_small_comps(image: NChanImg, minSzThreshold=30):
  areas, labels = _cvConnComps(image, areaOnly=True)
  validLabels = np.flatnonzero(areas >= minSzThreshold) + 1
  out = np.isin(labels, validLabels)
  return ProcessIO(image=out)

def draw_vertices(image: NChanImg, fgVerts: XYVertices):
  return ProcessIO(image=ComplexXYVertices([fgVerts]).toMask(image.shape[:2], asBool=True))

def convert_to_squares(image: NChanImg):
  outMask = np.zeros(image.shape, dtype=bool)
  for region in regionprops(label(image)):
    outMask[region.bbox[0]:region.bbox[2],region.bbox[1]:region.bbox[3]] = True
  return ProcessIO(image=outMask)

def _grabcutResultToMask(gcResult):
  return np.where((gcResult==2)|(gcResult==0), False, True)

class _CvGrabcut(AtomicProcess):


  def __init__(self, **kwargs):
    super().__init__(self.grabcut, 'Cv Grabcut', **kwargs)

  def grabcut(self, image: NChanImg, prevCompMask: BlackWhiteImg, fgVerts: XYVertices,
               firstRun: bool, historyMask: GrayImg, iters=5):
    if image.size == 0:
      return ProcessIO(image=np.zeros_like(prevCompMask))
    img = cv.cvtColor(image, cv.COLOR_RGB2BGR)
    historyMask = historyMask.copy()
    if historyMask.size:
      historyMask[fgVerts.rows, fgVerts.cols] = 2

    mask = np.zeros(prevCompMask.shape, dtype='uint8')
    if historyMask.shape == mask.shape:
      mask[prevCompMask == 1] = cv.GC_PR_FGD
      mask[prevCompMask == 0] = cv.GC_PR_BGD
      mask[historyMask == 2] = cv.GC_FGD
      mask[historyMask == 1] = cv.GC_BGD

    cvRect = np.array([fgVerts.min(0), fgVerts.max(0) - fgVerts.min(0)]).flatten()

    if firstRun or self.result is None or 'fgdModel' not in self.result:
      bgdModel = np.zeros((1, 65), np.float64)
      fgdModel = np.zeros((1, 65), np.float64)
    else:
      bgdModel = self.result['bgdModel']
      fgdModel = self.result['fgdModel']

    if not np.any(mask):
      if cvRect[2] == 0 or cvRect[3] == 0:
        return ProcessIO(image=np.zeros_like(prevCompMask))
      mode = cv.GC_INIT_WITH_RECT
    else:
      mode = cv.GC_INIT_WITH_MASK
    cv.grabCut(img, mask, cvRect, bgdModel, fgdModel, iters, mode=mode)
    outMask = np.where((mask==2)|(mask==0), False, True)
    return ProcessIO(labels=outMask, fgdModel=fgdModel, bgdModel=bgdModel)

def cv_grabcut_factory():
  return _CvGrabcut()

def quickshift_seg(image: NChanImg, ratio=1.0, max_dist=10.0, kernel_size=5,
                   sigma=0.0):
  # For max_dist of 0, the input isn't changed and it takes a long time
  key = (max_dist, kernel_size, sigma)
  if max_dist == 0:
    # Make sure output is still 1-channel
    segImg = image.mean(2).astype(int) if image.ndim > 2 else image
  else:
    if image.ndim < 3:
      image = np.tile(image[:,:,None], (1,1,3))
    segImg = seg.quickshift(image, ratio=ratio, kernel_size=kernel_size, max_dist=max_dist, sigma=sigma)
  return ProcessIO(labels=segImg)
quickshift_seg.__doc__ = seg.quickshift.__doc__

# Taken from example page: https://scikit-image.org/docs/dev/auto_examples/segmentation/plot_morphsnakes.html
def morph_acwe(image: NChanImg, initialCheckerSize=6, iters=35, smoothing=3):
  image = img_as_float(image)
  if image.ndim > 2:
    image = image.mean(2)
  initLs = seg.checkerboard_level_set(image.shape, initialCheckerSize)
  outLevels = seg.morphological_chan_vese(image, iters, init_level_set=initLs, smoothing=smoothing)
  return ProcessIO(labels=outLevels)

def k_means_segmentation(image: NChanImg, kVal=5, attempts=10):
  # Logic taken from https://docs.opencv.org/master/d1/d5c/tutorial_py_kmeans_opencv.html
  numChannels = 1 if image.ndim < 3 else image.shape[2]
  clrs = image.reshape(-1, numChannels)
  clrs = clrs.astype('float32')
  criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 10, 1.0)
  ret, lbls, imgMeans = cv.kmeans(clrs, kVal, None, criteria, attempts,
                             cv.KMEANS_RANDOM_CENTERS)
  # Now convert back into uint8, and make original image
  imgMeans = imgMeans.astype('uint8')
  lbls = lbls.reshape(image.shape[:2])

  return ProcessIO(labels=lbls, means=imgMeans)

def _labelBoundaries_cv(labels: np.ndarray, thickness: int):
  """Code stolen and reinterpreted for cv from skimage.segmentation.boundaries"""
  if thickness % 2 == 0:
    thickness += 1
  thickness = max(thickness, 3)
  if labels.dtype not in [np.uint8, np.uint16, np.int16, np.float16, np.float32]:
    labels = labels.astype(np.uint16)
  strel = cv.getStructuringElement(cv.MORPH_RECT, (thickness, thickness))
  return cv.morphologyEx(labels, cv.MORPH_DILATE, strel) \
               != cv.morphologyEx(labels, cv.MORPH_ERODE, strel)

def binarize_labels(image: NChanImg, labels: BlackWhiteImg, fgVerts: XYVertices,
                    historyMask: BlackWhiteImg, touchingRoiOnly=True, useMeanColor=True,
                    lineThickness=2):
  """
  For a given binary image input, only keeps connected components that are directly in
  contact with at least one of the specified vertices. In essence, this function can make
  a wide variety of operations behave similarly to region growing.

  :param touchingRoiOnly: Whether to only keep labeled regions that are in contact
    with the current ROI
  :param useMeanColor: Whether to color the summary info image with mean values or
    (if *False*) just draw the boundaries around each label.
  :param lineThickness:
    helpText: How thick to draw label boundary and ROI vertices lines
  """
  if labels.ndim > 2:
    raise ValueError('Cannot handle multichannel labels.\n'
                     f'(labelss.shape={labels.shape})')
  # seeds = cornersToFullBoundary(fgVerts, 50e3)[:, ::-1]
  seeds = np.clip(fgVerts[:, ::-1], 0, np.array(labels.shape)-1)
  if image.ndim < 3:
    image = image[...,None]
  if touchingRoiOnly:
    out = np.zeros_like(labels, dtype=bool)
    for seed in seeds:
      out |= flood(labels, tuple(seed),)
  else:
    keepColors = labels[seeds[:,0], seeds[:,1]]
    out = np.isin(labels, keepColors)
  # Zero out negative regions from previous runs
  if np.issubdtype(labels.dtype, np.bool_):
    # Done for stage summary only
    labels = label(labels)
  if historyMask.shape == out.shape:
    out[historyMask == 1] = False
  nChans = image.shape[2]
  if useMeanColor:
    summaryImg = np.zeros_like(image)
    # Offset by 1 to avoid missing 0-labels
    for lbl in regionprops(labels+1):
      coords = lbl.coords
      intensity = image[coords[:,0], coords[:,1],...].mean(0)
      summaryImg[coords[:,0], coords[:,1], :] = intensity
  else:
    boundaries = _labelBoundaries_cv(labels, lineThickness)
    summaryImg = image.copy()
    summaryImg[boundaries,...] = [255 for _ in range(nChans)]
  color = (255,) + tuple(0 for _ in range(1, nChans))
  cv.drawContours(summaryImg, [fgVerts], -1, color, lineThickness)
  return ProcessIO(image=out, summaryInfo={'image': summaryImg})

def region_grow_segmentation(image: NChanImg, fgVerts: XYVertices, seedThresh=10):
  if image.size == 0:
    return ProcessIO(image=np.zeros(image.shape[:2], bool))
  if np.all(fgVerts == fgVerts[0, :]):
    # Remove unnecessary redundant seedpoints
    fgVerts = fgVerts[[0], :]
  # outMask = np.zeros(image.shape[0:2], bool)
  # For small enough shapes, get all boundary pixels instead of just shape vertices
  if fgVerts.connected:
    fgVerts = cornersToFullBoundary(fgVerts, 50e3)

  # Don't let region grow outside area of effect
  # img_aoe, coords = getCroppedImg(image, fgVerts, areaOfEffect, coordsAsSlices=True)
  # Offset vertices before filling
  # seeds = fgVerts - [coords[1].start, coords[0].start]
  outMask = growSeedpoint(image, fgVerts, seedThresh)

  return ProcessIO(image=outMask)

def slic_segmentation(image, n_segments=100, compactness=10.0, sigma=0, min_size_factor=0.5, max_size_factor=3):
  return ProcessIO(labels=seg.slic(**locals(), start_label=1))
slic_segmentation.__doc__ = seg.slic.__doc__

@fns.dynamicDocstring(inters=[attr for attr in vars(cv) if attr.startswith('INTER_')])
def cv_resize(image: np.ndarray, newSize: Union[float, tuple]=0.5, asRatio=True, interpolation='INTER_CUBIC'):
  """
  :param image: Image to resize
  :param interpolation:
    pType: list
    limits: {inters}
  :param newSize:
    pType: float
    step: 0.1
  :param asRatio:
    readonly: True
  """
  if isinstance(interpolation, str):
    interpolation = getattr(cv, interpolation)
  return tryCvResize(image, newSize, asRatio, interpolation)


from argparse import Action

from pyqtgraph.Qt import QtCore

from s3a.constants import PRJ_ENUMS
from utilitys import fns
from . import __version__
from . import appInst
from .views.s3agui import S3A


def main(loadLastState=True, **load):
  """
  Calling code for the S3A application.

  :param loadLastState: When the app is closed, all settings are saved. If this is *True*,
    these settings are restored on startup. If *False*, they aren't.
  :param load: States to load, see the help output for possible values
  """
  # Handle here for faster bootup
  win = S3A(log=PRJ_ENUMS.LOG_GUI, loadLastState=loadLastState, **load)
  QtCore.QTimer.singleShot(0, win.showMaximized)
  try:
    appInst.exec_()
  except AttributeError:
    appInst.exec()
  return win

def main_cli():
  parser = fns.makeCli(main, parserKwargs=dict(prog='S3A', add_help=False))
  parser.register('action', 'help', S3AHelp)
  parser.add_argument('--version', action='version', version=__version__)
  parser.add_argument('--help', action='help')
  args = parser.parse_args()
  main(**vars(args))

class S3AHelp(Action):
  def __init__(self, **kwargs):
    kwargs.update(nargs=0)
    super().__init__(**kwargs)

  def __call__(self, parser, *args, **kwargs) -> None:
    win = S3A(loadLastState=False, log=PRJ_ENUMS.LOG_NONE)
    win.makeHelpOpts(parser)
    # Prevent settings overwrite
    win.appStateEditor.saveParamValues = lambda *args, **kwargs: None
    win.forceClose()
    parser.print_help()
    parser.exit()


if __name__ == '__main__':
  main_cli()


import warnings
from collections import defaultdict
from pathlib import Path
from typing import Union, Dict, List, Sequence

import pyqtgraph as pg
import qdarkstyle
from pandas import DataFrame as df
from pyqtgraph.Qt import QtCore, QtWidgets, QtGui

from s3a.constants import LAYOUTS_DIR, REQD_TBL_FIELDS, ICON_DIR
from s3a.constants import PRJ_ENUMS, PRJ_CONSTS
from s3a.generalutils import hierarchicalUpdate
from s3a.models.s3abase import S3ABase
from s3a.plugins.misc import RandomToolsPlugin, MainImagePlugin, CompTablePlugin
from s3a.shared import SharedAppSettings
from s3a.structures import FilePath, NChanImg
from utilitys import ParamEditor, ParamEditorPlugin, RunOpts, PrjParam, fns

__all__ = ['S3A']

_MENU_PLUGINS = [RandomToolsPlugin]

class S3A(S3ABase):
  sigLayoutSaved = QtCore.Signal()

  __groupingName__ = 'Main Window'

  def __initEditorParams__(self, shared: SharedAppSettings):
    super().__initEditorParams__(shared)
    self.toolsEditor = ParamEditor.buildClsToolsEditor(type(self), 'Main Window')
    shared.colorScheme.registerFunc(self.updateTheme, runOpts=RunOpts.ON_CHANGED, nest=False)


  def __init__(self, parent=None, log: Union[str, Sequence[str]]=PRJ_ENUMS.LOG_TERM,
               loadLastState=True, **startupSettings):
    # Wait to import quick loader profiles until after self initialization so
    # customized loading functions also get called
    super().__init__(parent, **startupSettings)
    self.setWindowIcon(QtGui.QIcon(str(ICON_DIR/'s3alogo.svg')))
    if PRJ_ENUMS.LOG_GUI in log:
      warnings.simplefilter('error', UserWarning)
      fns.makeExceptionsShowDialogs(self)
    self.APP_TITLE = 'FICS Semi-Supervised Semantic Annotator'
    self.CUR_COMP_LBL = 'Current Component ID:'
    self.setWindowTitle(self.APP_TITLE)
    self.setWindowIconText(self.APP_TITLE)

    self.curCompIdLbl = QtWidgets.QLabel(self.CUR_COMP_LBL)

    # -----
    # LAOYUT MANAGER
    # -----
    # Dummy editor for layout options since it doesn't really have editable settings
    # Maybe later this can be elevated to have more options
    self.layoutEditor = ParamEditor(self, None, LAYOUTS_DIR, 'dockstate', 'Layout')
    def loadLayout(layoutName: Union[str, Path]):
      layoutName = Path(layoutName)
      if not layoutName.is_absolute():
        layoutName = LAYOUTS_DIR/f'{layoutName}.dockstate'
      state = fns.attemptFileLoad(layoutName)
      self.restoreState(state['docks'])

    def saveRecentLayout(_folderName: Path):
      outFile = _folderName/'layout.dockstate'
      self.saveLayout(outFile)
      return str(outFile)

    self.layoutEditor.loadParamValues = loadLayout
    self.layoutEditor.saveParamValues = saveRecentLayout
    self.appStateEditor.addImportExportOpts('layout', loadLayout, saveRecentLayout)

    self._buildGui()
    self._buildMenu()
    self._hookupSignals()

    # Load in startup settings
    stateDict = None if loadLastState else {}
    hierarchicalUpdate(self.appStateEditor.startupSettings, startupSettings)
    self.appStateEditor.loadParamValues(stateDict=stateDict)

  def _hookupSignals(self):
    # EDIT
    self.saveAllEditorDefaults()

  def _buildMenu(self):
    # Nothing to do for now
    pass

  def _buildGui(self):
    self.setDockOptions(self.ForceTabbedDocks)
    self.setTabPosition(QtCore.Qt.AllDockWidgetAreas, QtWidgets.QTabWidget.North)
    centralWidget = QtWidgets.QWidget()
    self.setCentralWidget(centralWidget)
    layout = QtWidgets.QVBoxLayout(centralWidget)

    self.toolbarWidgets: Dict[PrjParam, List[QtWidgets.QAction]] = defaultdict(list)
    layout.addWidget(self.mainImg)

    self.tblFieldToolbar.setObjectName('Table Field Plugins')
    # self.addToolBar(self.tblFieldToolbar)
    self.tblFieldToolbar.hide()
    self.generalToolbar.setObjectName('General')
    self.addToolBar(self.generalToolbar)

    _plugins = [self.clsToPluginMapping[c] for c in [MainImagePlugin, CompTablePlugin]]
    parents = [self.mainImg, self.compTbl]
    for plugin, parent in zip(_plugins, reversed(parents)):
      plugin.toolsEditor.actionsMenuFromProcs(plugin.name, nest=True, parent=parent, outerMenu=parent.menu)


    tableDock = QtWidgets.QDockWidget('Component Table Window', self)
    tableDock.setFeatures(tableDock.DockWidgetMovable|tableDock.DockWidgetFloatable)

    tableDock.setObjectName('Component Table Dock')
    tableContents = QtWidgets.QWidget(tableDock)
    tableLayout = QtWidgets.QVBoxLayout(tableContents)
    tableLayout.addWidget(self.compTbl)
    tableDock.setWidget(tableContents)

    self.addDockWidget(QtCore.Qt.BottomDockWidgetArea, tableDock)

    # STATUS BAR
    self.setStatusBar(self.statBar)

    self.mouseCoordsLbl = QtWidgets.QLabel()
    self.pxColorLbl = QtWidgets.QLabel()

    self.imageLbl = QtWidgets.QLabel(f"Image: None")

    self.statBar.show()
    self.statBar.addWidget(self.imageLbl)

    self.statBar.addWidget(self.mouseCoordsLbl)
    self.mainImg.mouseCoordsLbl = self.mouseCoordsLbl

    self.statBar.addWidget(self.pxColorLbl)
    self.mainImg.pxColorLbl = self.pxColorLbl

  def saveLayout(self, layoutName: Union[str, Path]=None, allowOverwriteDefault=False):
    dockStates = self.saveState().data()
    if Path(layoutName).is_absolute():
      savePathPlusStem = layoutName
    else:
      savePathPlusStem = LAYOUTS_DIR/layoutName
    saveFile = savePathPlusStem.with_suffix(f'.dockstate')
    fns.saveToFile({'docks': dockStates}, saveFile,
                   allowOverwriteDefault=allowOverwriteDefault)
    self.sigLayoutSaved.emit()

  def changeFocusedComp(self, compIds: Union[int, Sequence[int]]=None):
    ret = super().changeFocusedComp(compIds)
    self.curCompIdLbl.setText(f'Component ID: {self.mainImg.compSer[REQD_TBL_FIELDS.INST_ID]}')
    return ret

  def resetTblFields_gui(self):
    outFname = fns.popupFilePicker(None, 'Select Table Config File', 'All Files (*.*);; Config Files (*.yml)')
    if outFname is not None:
      self.sharedAttrs.tableData.loadCfg(outFname)

  def _addPluginObj(self, plugin: ParamEditorPlugin, **kwargs):
    plugin = super()._addPluginObj(plugin, **kwargs)
    if not plugin:
      return
    dock = plugin.dock
    if dock is None:
      return
    self.sharedAttrs.quickLoader.addDock(dock)
    self.addTabbedDock(QtCore.Qt.RightDockWidgetArea, dock)

    if plugin.menu is None:
      # No need to add menu and graphics options
      return plugin

    parentTb = plugin.parentMenu
    if parentTb is not None:
      plugin.addToWindow(self, parentToolbarOrMenu=parentTb)
    return plugin

  def setMainImg(self, fileName: FilePath = None, imgData: NChanImg = None,
                 clearExistingComps=True):
    gen = super().setMainImg(fileName, imgData, clearExistingComps)
    ret = fns.gracefulNext(gen)
    img = self.srcImgFname
    if img is not None:
      img = img.name
    self.imageLbl.setText(f'Image: {img}')

    yield ret
    yield fns.gracefulNext(gen)

  def setMainImg_gui(self):
    fileFilter = "Image Files (*.png *.tif *.jpg *.jpeg *.bmp *.jfif);;All files(*.*)"
    fname = fns.popupFilePicker(None, 'Select Main Image', fileFilter)
    if fname is not None:
      with pg.BusyCursor():
        self.setMainImg(fname)

  def exportAnnotations_gui(self):
    """Saves the component table to a file"""
    fileFilters = self.compIo.ioFileFilter(**{'*': 'All Files'})
    outFname = fns.popupFilePicker(None, 'Select Save File', fileFilters, existing=False)
    if outFname is not None:
      super().exportCurAnnotation(outFname)

  def openAnnotation_gui(self):
    # TODO: See note about exporting comps. Delegate the filepicker activity to importer
    fileFilter = self.compIo.ioFileFilter(which=PRJ_ENUMS.IO_IMPORT)
    fname = fns.popupFilePicker(None, 'Select Load File', fileFilter)
    if fname is None:
      return
    self.openAnnotations(fname)

  def saveLayout_gui(self):
    outName = fns.dialogGetSaveFileName(self, 'Layout Name')
    if outName is None or outName == '':
      return
    self.saveLayout(outName)

  # ---------------
  # BUTTON CALLBACKS
  # ---------------
  def closeEvent(self, ev: QtGui.QCloseEvent):
    # Confirm all components have been saved
    shouldExit = True
    forceClose = False
    if self.hasUnsavedChanges:
      ev.ignore()
      forceClose = False
      msg = QtWidgets.QMessageBox()
      msg.setWindowTitle('Confirm Exit')
      msg.setText('Component table has unsaved changes.\nYou can choose to save and exit or discard changes')
      msg.setDefaultButton(msg.Save)
      msg.setStandardButtons(msg.Discard|msg.Cancel|msg.Save)
      code = msg.exec_()
      if code == msg.Discard:
        forceClose = True
      elif code == msg.Cancel:
        shouldExit = False
    if shouldExit:
      # Clean up all editor windows, which could potentially be left open
      ev.accept()
      fns.restoreExceptionBehavior()
      if not forceClose:
        self.appStateEditor.saveParamValues()

  def forceClose(self):
    """
    Allows the app to close even if it has unsaved changes. Useful for closing
    within a script
    """
    self.hasUnsavedChanges = False
    self.close()

  def _populateLoadLayoutOptions(self):
    self.layoutEditor.addDirItemsToMenu(self.menuLayout)

  def updateTheme(self, useDarkTheme=False):
    style = ''
    if useDarkTheme:
      style = qdarkstyle.load_stylesheet()
    self.setStyleSheet(style)

  def add_focusComps(self, newComps: df, addType=PRJ_ENUMS.COMP_ADD_AS_NEW):
    ret = super().add_focusComps(newComps, addType=addType)
    selection = self.compDisplay.selectRowsById(newComps[REQD_TBL_FIELDS.INST_ID])
    if self.isVisible() and self.compTbl.props[PRJ_CONSTS.PROP_SHOW_TBL_ON_COMP_CREATE]:
      # For some reason sometimes the actual table selection doesn't propagate in time, so
      # directly forward the selection here
      self.compTbl.setSelectedCellsAs_gui(selection)
    return ret

if __name__ == '__main__':
  import sys
  app = QtWidgets.QApplication([])
  win = S3A()
  win.showMaximized()
  sys.exit(app.exec_())


from pyqtgraph.Qt import QtCore

from s3a.constants import GEN_PROPS_DIR, SCHEMES_DIR, SHORTCUTS_DIR, IMG_PROC_DIR, \
  MULT_PRED_DIR, CFG_DIR
from s3a.parameditors.algcollection import AlgCollection
from s3a.parameditors.quickloader import QuickLoaderEditor
from s3a.parameditors.table import TableData
from s3a.plugins.misc import SettingsPlugin, ShortcutsPlugin
from s3a.processing import ImgProcWrapper, ImageProcess
from utilitys import ActionStack, ParamEditor, ShortcutParameter, dockPluginFactory


class SharedAppSettings(QtCore.QObject):
  def __init__(self, parent=None):
    super().__init__(parent)
    self.actionStack = ActionStack()

    self.tableData = TableData()
    self.filter = self.tableData.filter

    self.imgProcClctn = AlgCollection(ImgProcWrapper, ImageProcess, saveDir=IMG_PROC_DIR,
                                      template=CFG_DIR/'imageproc.yml')
    self.multiPredClctn = AlgCollection(saveDir=MULT_PRED_DIR, template=CFG_DIR/'multipred.yml')

    self.settingsPlg = SettingsPlugin()
    self.colorScheme = self.settingsPlg.colorScheme
    self.generalProps = self.settingsPlg.generalProps

    self.shortcutsPlg = ShortcutsPlugin()
    self.shortcuts = self.shortcutsPlg.shortcuts
    self.quickLoader = self.shortcutsPlg.quickLoader



from __future__ import annotations

from ast import literal_eval
from typing import Union, List, Sequence
from warnings import warn

import cv2 as cv
import numpy as np

from .typeoverloads import NChanImg, BlackWhiteImg

class XYVertices(np.ndarray):
  connected = True

  def __new__(cls, inputArr: Union[list, np.ndarray, tuple]=None, connected=True, dtype=int,
              **kwargs):
    # Default to integer type if not specified, since this is how pixel coordinates will be represented anyway
    # See numpy docs on subclassing ndarray
    if inputArr is None:
      inputArr = np.zeros((0,2))
    arr = np.asarray(inputArr, dtype=dtype, **kwargs).view(cls)
    arr.connected = connected
    return arr

  def __array_finalize__(self, obj):
    shape = self.shape
    shapeLen = len(shape)
    # indicates point, so the one dimension must have only 2 elements
    if 1 < shapeLen < 2 and shape[0] != 2:
      raise ValueError(f'A one-dimensional vertex array must be shape (2,).'
                                f' Receieved array of shape {shape}')
    elif shapeLen > 2 or shapeLen > 1 and shape[1] != 2:
      raise ValueError(f'Vertex list must be Nx2. Received shape {shape}.')
    if obj is None: return
    self.connected = getattr(obj, 'connected', True)

  @property
  def empty(self):
      return len(self) == 0

  def serialize(self):
    return str(self.tolist())

  @classmethod
  def deserialize(cls, strVal: str):
    out = cls(literal_eval(strVal))
    if out.size == 0:
      # Make sure size is 0x2
      return cls()
    return out

  # def asPoint(self):
  #   if self.size == 2:
  #     return self.reshape(-1)
  #   # Reaching here means the user requested vertices as point when
  #   # more than one point is in the list
  #   raise ValueError(f'asPoint() can only be called when one vertex is in'
  #                             f' the vertex list. Currently has shape {self.shape}')

  # def asRowCol(self):
  #   return np.fliplr(self)

  @property
  def x(self):
    # Copy to array first so dimensionality checks are no longer required
    return np.array(self).reshape(-1,2)[:,[0]]
  @x.setter
  def x(self, newX): self.reshape(-1,2)[:,0] = newX

  @property
  def y(self):
    return np.array(self).reshape(-1,2)[:,[1]]
  @y.setter
  def y(self, newY): self.reshape(-1,2)[:,1] = newY

  @property
  def rows(self): return self.y
  @rows.setter
  def rows(self, newRows): self.y = newRows

  @property
  def cols(self):return self.x
  @cols.setter
  def cols(self, newCols):self.x = newCols


class ComplexXYVertices(list):
  """
  Allows holes in the component shape. Subclassing ndarray instead of list allows primitive algebraic ops on the list
  contents (e.g. subtracting/adding offset). Since normal usage doesn't typically require a mutable structure, the
  loss is minimal.
  """
  hierarchy = np.ones((0,4), dtype=int)
  """See cv.findContours for hierarchy explanation. Used in cv.RETR_CCOMP mode."""

  def __init__(self, inputArr: Union[List[XYVertices], np.ndarray, Sequence]=None,
               hierarchy:np.ndarray=None,
               coerceListElements=False):

    if inputArr is None:
      inputArr = []
    numInpts = len(inputArr)
    if coerceListElements:
      inputArr = [XYVertices(el) for el in inputArr]
    if hierarchy is None:
      hierarchy = np.ones((numInpts, 4), dtype=int)*-1
    super().__init__(inputArr)
    # No hierarchy required unless list is longer than length 1
    self.hierarchy = hierarchy

  def append(self, verts:XYVertices=None) -> None:
    if verts is not None:
      super().append(verts)

  def isEmpty(self):
    return len(self.stack()) == 0

  @property
  def x(self):
    return [lst.x for lst in self]
  @x.setter
  def x(self, newX):
    for lst, newLstX in zip(self, newX):
      lst.x = newLstX

  @property
  def y(self):
    return [lst.y for lst in self]

  @y.setter
  def y(self, newY):
    for lst, newLstY in zip(self, newY):
      lst.y = newLstY

  def asPoint(self):
    if len(self) == 1 and self[0].shape[0] == 1:
      return self[0][0]
    else:
      raise ValueError(f'Can only treat ComplexXYVertices with one inner list as a point.'
                       f' Current list has {len(self)} element(s), '
                       f'where element 0 has shape {self[0].shape}.')

  def stack(self, newDtype=None) -> XYVertices:
    if len(self) == 0:
      # Check required for np vstack since it won't work with a 0-element array
      return XYVertices()
    else:
      if newDtype is None:
        newDtype = self[0].dtype
      return XYVertices(np.vstack(self), dtype=newDtype)

  def removeOffset(self, offset=None, inplace=False, returnOffset=False):
    """
    Subtracts a constant offset from all contained vertices. If offset is None, it will be set to the min value across
    all contained vertices such that this polygon's coordinates are all relative to (0,0).
    """
    if offset is None:
      offset = self.stack().min(0)
    if inplace:
      for verts in self:
        verts -= offset
      if returnOffset:
        return offset
    else:
      out = ComplexXYVertices([verts - offset for verts in self], self.hierarchy)
      if returnOffset:
        return out, offset
      return out

  @classmethod
  def stackedMax(cls, complexVertList: Sequence[ComplexXYVertices]):
    """
    Returns the max along dimension 0 for a list of complex vertices
    """
    return np.vstack([v.stack() for v in complexVertList]).max(0)

  @classmethod
  def stackedMin(cls, complexVertList: Sequence[ComplexXYVertices]):
    return np.vstack([v.stack() for v in complexVertList]).min(0)


  def filledVerts(self) -> ComplexXYVertices:
    """
    Retrieves all vertex lists corresponding to filled regions in the complex shape
    """
    idxs = np.nonzero(self.hierarchy[:,3] == -1)[0]
    return ComplexXYVertices([self[ii] for ii in idxs])

  def holeVerts(self) -> ComplexXYVertices:
    """
    Retrieves all vertex lists corresponding to holes in the complex shape
    """
    idxs = np.nonzero(self.hierarchy[:,3] != -1)[0]
    return ComplexXYVertices([self[ii] for ii in idxs])

  def toMask(self, maskShape: Union[Sequence, NChanImg]=None,
             fillColor: Union[int, float, np.ndarray]=None,
             asBool=True, checkForDisconnectedVerts=False, warnIfTooSmall=True):
    if maskShape is None:
      try:
        maskShape = tuple(self.stack().max(0)[::-1]+1)
      except ValueError:
        # Mask is zero-sized
        dtype = 'bool' if asBool else 'uint16'
        return np.zeros((0,0), dtype)
      # Guaranteed not to be too small
      warnIfTooSmall = False
    if warnIfTooSmall:
      cmpShape = maskShape if isinstance(maskShape, Sequence) else maskShape.shape[:2]
      # Wait until inside 'if' so max isn't unnecessarily calculated
      # Edge case: Empty vertices set will trigger a value warning
      if len(self) == 0:
        vertMax = 0
      else:
        vertMax = self.stack().max(0)[::-1]
      if np.any(vertMax > np.array(cmpShape[:2])):
        warn('Vertices don\'t fit in the provided mask size.\n'
             f'Vertex shape: {vertMax}, mask shape: {cmpShape}')
    if checkForDisconnectedVerts:
      fillArg = []
      for verts in self: # type: XYVertices
        if verts.connected:
          fillArg.append(verts)
        else:
          # Make sure each point is treated separately, not part of a shape
          # to fill
          fillArg.extend(verts)
    else:
      fillArg = self
    if isinstance(maskShape, NChanImg):
      out = maskShape
    else:
      out = np.zeros(maskShape, 'uint16')
    nChans = 1 if out.ndim < 3 else out.shape[2]
    if fillColor is None:
      fillColor = tuple([1 for _ in range(nChans)])
    fillColorCmp = np.array(fillColor)
    if np.any((np.iinfo(out.dtype).min > fillColorCmp)
              | (fillColorCmp > np.iinfo(out.dtype).max)):
      raise ValueError('Fill color is larger or smaller than mask range can represent')
    cv.fillPoly(out, fillArg, fillColor)
    if asBool:
      return out > 0
    else:
      return out

  @staticmethod
  def fromBwMask(bwMask: BlackWhiteImg, simplifyVerts=True, externOnly=False) -> ComplexXYVertices:
    approxMethod = cv.CHAIN_APPROX_SIMPLE
    if not simplifyVerts:
      approxMethod = cv.CHAIN_APPROX_NONE
    retrMethod = cv.RETR_CCOMP
    if externOnly:
      retrMethod = cv.RETR_EXTERNAL
    # Contours are on the inside of components, so dilate first to make sure they are on the
    # outside
    #bwmask = dilation(bwmask, np.ones((3,3), dtype=bool))
    if bwMask.dtype != np.uint8:
      bwMask = bwMask.astype('uint8')
    contours, hierarchy = cv.findContours(bwMask, retrMethod, approxMethod)
    compVertices = ComplexXYVertices()
    for contour in contours:
      compVertices.append(XYVertices(contour[:,0,:]))
    if hierarchy is None:
      hierarchy = np.ones((0,1,4), int)*-1
    else:
      hierarchy = hierarchy[0,:,:]
    compVertices.hierarchy = hierarchy
    return compVertices

  def __str__(self) -> str:
    """
    Improve the readability of vertex list in table by just displaying stats of larger arrays
    :return: Human readable string representation
    """
    concatVerts = self.stack()
    if len(concatVerts) <= 4: return str(concatVerts)
    return f'Mean:\t{np.round(concatVerts.mean(0), 1)}\n' \
           f'# Points:\t{len(concatVerts)}\n' \
           f'Min:\t{concatVerts.min(0)}\n' \
           f'Max:\t{concatVerts.max(0)}'

  def __eq__(self, other: ComplexXYVertices):
    if len(self) != len(other):
      return False
    for selfVerts, otherVerts in zip(self, other):
      if not np.array_equal(selfVerts, otherVerts):
        return False
    return True

  def __ne__(self, other):
    return not self == other

  def copy(self) -> ComplexXYVertices:
    return ComplexXYVertices([lst.copy() for lst in self], self.hierarchy)

  def serialize(self):
    return str([arr.tolist() for arr in self])

  @staticmethod
  def deserialize(strObj: str) -> ComplexXYVertices:
    # TODO: Infer appropriate hierarchy from the serialized string. It is possible by finding whether vertices are given
    #  in CW or CCW order. This doesn't affect how they are drawn, but it does effect the return values of "holeVerts()"
    #  and "filledVerts()"
    outerLst = literal_eval(strObj)
    return ComplexXYVertices([XYVertices(lst) for lst in outerLst])